# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests for optimizers."""

from torch import Tensor, manual_seed
from torch.nn import Linear
from torch.optim import SGD, AdamW
from torch.nn.functional import mse_loss

from aihwkit.optim import AnalogOptimizer

from .helpers.testcases import AihwkitTestCase


class InferenceOptimizerTest(AihwkitTestCase):
    """Tests for the AnalogOptimizer."""

    @staticmethod
    def train_model(model, opt, loss_func, x_b, y_b):
        """Train the model."""
        epochs = 100
        for _ in range(epochs):
            opt.zero_grad()
            pred = model(x_b)
            loss = loss_func(pred, y_b)

            loss.backward()
            opt.step()

        return loss

    def test_instantiate(self):
        """Test instantiating the optimizer."""
        model = Linear(3, 4)
        optimizer = AnalogOptimizer(SGD, model.parameters(), lr=0.123)

        # Assert that a new subclass is created, with both the wrapped
        # optimizer and the AnalogInferenceOptimizer as parents.
        self.assertIsInstance(optimizer, AnalogOptimizer)
        self.assertIsInstance(optimizer, SGD)
        self.assertIsNot(type(optimizer), AnalogOptimizer)

        # Assert over specific wrapped optimizer parameters.
        self.assertEqual(optimizer.defaults['lr'], 0.123)

    def test_train_digital_sgd(self):
        """Test training digital layer with analog optimizer (SGD)."""
        loss_func = mse_loss

        x_b = Tensor([[0.1, 0.2, 0.3, 0.1], [0.2, 0.4, 0.3, 0.6]])
        y_b = Tensor([[0.3], [0.6]])

        manual_seed(4321)
        model1 = Linear(4, 1)

        manual_seed(4321)
        model2 = Linear(4, 1)

        learning_rate = 0.123
        optimizer1 = AnalogOptimizer(SGD, model1.parameters(), lr=learning_rate)
        optimizer2 = SGD(model2.parameters(), lr=learning_rate)

        loss1 = self.train_model(model1, optimizer1, loss_func, x_b, y_b)
        loss2 = self.train_model(model2, optimizer2, loss_func, x_b, y_b)

        self.assertAlmostEqual(loss1, loss2)
        self.assertTensorAlmostEqual(model1.weight, model2.weight)

    def test_train_digital_adamw(self):
        """Test training digital layer with analog optimizer (AdamW)."""
        loss_func = mse_loss

        x_b = Tensor([[0.1, 0.2, 0.3, 0.1], [0.2, 0.4, 0.3, 0.6]])
        y_b = Tensor([[0.3], [0.6]])

        manual_seed(4321)
        model1 = Linear(4, 1)

        manual_seed(4321)
        model2 = Linear(4, 1)

        learning_rate = 0.123
        optimizer1 = AnalogOptimizer(AdamW, model1.parameters(), lr=learning_rate)
        optimizer2 = AdamW(model2.parameters(), lr=learning_rate)

        loss1 = self.train_model(model1, optimizer1, loss_func, x_b, y_b)
        loss2 = self.train_model(model2, optimizer2, loss_func, x_b, y_b)

        self.assertAlmostEqual(loss1, loss2)
        self.assertTensorAlmostEqual(model1.weight, model2.weight)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Data classes for the AIHW Composer API."""

from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import Any, Optional

from aihwkit.cloud.client.exceptions import ExperimentStatusError
from aihwkit.cloud.converter.definitions.input_file_pb2 import TrainingInput
from aihwkit.cloud.converter.definitions.output_file_pb2 import TrainingOutput
from aihwkit.cloud.converter.v1.training import BasicTrainingConverter, BasicTrainingResultConverter
from aihwkit.experiments import BasicTraining


class CloudJobStatus(Enum):
    """Status for a CloudJob."""

    UNKNOWN = 0
    WAITING = 1
    RUNNING = 2
    COMPLETED = 3
    FAILED = 4


class CloudExperimentCategory(Enum):
    """Category of a CloudExperiment."""

    BASIC_TRAINING = 1


@dataclass
class CloudJob:
    """Representation of an API CloudJob."""

    id_: str
    output_id: Optional[str] = field(repr=False)
    status: CloudJobStatus = CloudJobStatus.UNKNOWN


@dataclass
class CloudExperiment:
    """Representation of an API Experiment."""

    _api_client: Any = field(repr=False)
    id_: str
    name: str
    category: CloudExperimentCategory = field(repr=False)
    created_at: datetime = field(repr=False)
    input_id: Optional[str] = field(repr=False)
    job: Optional[CloudJob] = field(repr=False)

    def get_experiment(self) -> BasicTraining:
        """Return a data Experiment.

        Returns:
            The experiment.

        Raises:
            ExperimentStatusError: if the Experiment is not in a valid status.
        """
        if self.status() == CloudJobStatus.UNKNOWN:
            raise ExperimentStatusError('Experiment input is not available')

        input_ = self._api_client.input_get(self.input_id)

        input_proto = TrainingInput()
        input_proto.ParseFromString(input_)

        converter = BasicTrainingConverter()
        return converter.from_proto(input_proto)

    def get_result(self) -> list:
        """Return the result of an Experiment.

        Returns:
            The experiment result.

        Raises:
            ExperimentStatusError: if the Experiment is not completed.
        """
        if self.status() != CloudJobStatus.COMPLETED:
            raise ExperimentStatusError('Output cannot be retrieved unless the '
                                        'experiment is completed')

        # Fetch the protobuf output.
        output_ = self._api_client.output_get(self.job.output_id)  # type: ignore

        # Convert from protobuf.
        training_output = TrainingOutput()
        training_output.ParseFromString(output_)
        converter = BasicTrainingResultConverter()
        output = converter.from_proto(training_output)

        return output['epochs']

    def status(self) -> CloudJobStatus:
        """Return the status of the experiment."""
        # Populate the job if not present.
        if not self.job:
            tmp_experiment = self._api_client.experiment_get(self.id_)
            self.input_id = tmp_experiment.input_id
            self.job = tmp_experiment.job

        # Fallback for Experiments without Job.
        if not self.job:
            return CloudJobStatus.UNKNOWN

        # Avoid refreshing statuses of Jobs in end states.
        if self.job.status == CloudJobStatus.COMPLETED and self.job.output_id:
            return CloudJobStatus.COMPLETED
        if self.job.status == CloudJobStatus.FAILED:
            return CloudJobStatus.FAILED

        # Refresh the status.
        self.job = self._api_client.job_get(self.job.id_)
        return self.job.status  # type: ignore



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests for analog presets."""

from torch import Tensor

from aihwkit.simulator.tiles.analog import AnalogTile
from aihwkit.simulator.presets import (
    ReRamESPreset, ReRamSBPreset, CapacitorPreset, EcRamPreset, EcRamMOPreset, IdealizedPreset,
    GokmenVlasovPreset, PCMPreset,
    ReRamES2Preset, ReRamSB2Preset, Capacitor2Preset, EcRam2Preset, EcRamMO2Preset,
    Idealized2Preset, ReRamES4Preset, ReRamSB4Preset, Capacitor4Preset, EcRam4Preset,
    EcRamMO4Preset, Idealized4Preset, TikiTakaReRamESPreset, TikiTakaReRamSBPreset,
    TikiTakaCapacitorPreset, TikiTakaEcRamPreset, TikiTakaEcRamMOPreset, TikiTakaIdealizedPreset,
    MixedPrecisionReRamESPreset, MixedPrecisionReRamSBPreset, MixedPrecisionCapacitorPreset,
    MixedPrecisionEcRamPreset, MixedPrecisionEcRamMOPreset, MixedPrecisionIdealizedPreset,
    MixedPrecisionGokmenVlasovPreset, MixedPrecisionPCMPreset,
)
from .helpers.decorators import parametrize_over_presets
from .helpers.testcases import AihwkitTestCase


@parametrize_over_presets(
    [
        ReRamESPreset, ReRamSBPreset, CapacitorPreset, EcRamPreset, EcRamMOPreset, IdealizedPreset,
        GokmenVlasovPreset, PCMPreset,
        ReRamES2Preset, ReRamSB2Preset, Capacitor2Preset, EcRam2Preset, EcRamMO2Preset,
        Idealized2Preset, ReRamES4Preset, ReRamSB4Preset, Capacitor4Preset, EcRam4Preset,
        EcRamMO4Preset, Idealized4Preset, TikiTakaReRamESPreset, TikiTakaReRamSBPreset,
        TikiTakaCapacitorPreset, TikiTakaEcRamPreset, TikiTakaEcRamMOPreset,
        TikiTakaIdealizedPreset,
        MixedPrecisionReRamESPreset, MixedPrecisionReRamSBPreset, MixedPrecisionCapacitorPreset,
        MixedPrecisionEcRamPreset, MixedPrecisionEcRamMOPreset, MixedPrecisionIdealizedPreset,
        MixedPrecisionGokmenVlasovPreset, MixedPrecisionPCMPreset,
    ]
)
class PresetTest(AihwkitTestCase):
    """Test for analog presets."""

    def test_tile_preset(self):
        """Test instantiating a tile that uses a preset."""
        out_size = 2
        in_size = 3

        rpu_config = self.preset_cls()
        analog_tile = AnalogTile(out_size, in_size, rpu_config, bias=False)

        learning_rate = 0.123

        # Use [out_size, in_size] weights.
        weights = Tensor([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]])

        # Set some properties in the simulators.Tile.
        analog_tile.set_learning_rate(0.123)
        analog_tile.set_weights(weights)

        # Assert over learning rate.
        self.assertAlmostEqual(analog_tile.get_learning_rate(), learning_rate)
        self.assertAlmostEqual(analog_tile.get_learning_rate(),
                               analog_tile.tile.get_learning_rate())

        # Assert over weights and biases.
        tile_weights, tile_biases = analog_tile.get_weights()
        self.assertEqual(tuple(tile_weights.shape), (out_size, in_size))
        self.assertEqual(tile_biases, None)
        # TODO: disabled as the comparison needs to take into account noise
        # self.assertTensorAlmostEqual(tile_weights, weights)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Analog layers."""

from typing import Optional

from torch import Tensor
from torch.nn import Linear

from aihwkit.nn.functions import AnalogFunction
from aihwkit.nn.modules.base import AnalogModuleBase, RPUConfigAlias


class AnalogLinear(AnalogModuleBase, Linear):
    """Linear layer that uses an analog tile.

    Linear layer that uses an analog tile during its forward, backward and
    update passes.

    Note:
        The tensor parameters of this layer (``.weight`` and ``.bias``) are not
        guaranteed to contain the same values as the internal weights and biases
        stored in the analog tile. Please use ``set_weights`` and
        ``get_weights`` when attempting to read or modify the weight/bias. This
        read/write process can simulate the (noisy and inexact) analog writing
        and reading of the resistive elements.

    Args:
        in_features: input vector size (number of columns).
        out_features: output vector size (number of rows).
        rpu_config: resistive processing unit configuration.
        bias: whether to use a bias row on the analog tile or not
        realistic_read_write: whether to enable realistic read/write
            for setting initial weights and read out of weights
        weight_scaling_omega: the weight value where the max
            weight will be scaled to. If zero, no weight scaling will
            be performed
    """
    # pylint: disable=abstract-method

    __constants__ = ['in_features', 'out_features']
    in_features: int
    out_features: int
    realistic_read_write: bool
    weight_scaling_omega: float

    def __init__(
            self,
            in_features: int,
            out_features: int,
            bias: bool = True,
            rpu_config: Optional[RPUConfigAlias] = None,
            realistic_read_write: bool = False,
            weight_scaling_omega: float = 0.0,
    ):
        # Create the tile.
        self.analog_tile = self._setup_tile(in_features,
                                            out_features,
                                            bias,
                                            rpu_config,
                                            realistic_read_write,
                                            weight_scaling_omega)
        # Call super() after tile creation, including ``reset_parameters``.
        super().__init__(in_features, out_features, bias=bias)

        # Unregister weight/bias as a parameter but keep it as a
        # field (needed for syncing still)
        self.unregister_parameter('weight')
        if bias:
            self.unregister_parameter('bias')

        # Register tile instead
        self.register_analog_tile(self.analog_tile)

    @classmethod
    def from_digital(
            cls,
            module: Linear,
            rpu_config: Optional[RPUConfigAlias] = None,
            realistic_read_write: bool = False,
            weight_scaling_omega: float = 0.0,
    ) -> 'AnalogLinear':
        """Return an AnalogLinear layer from a torch Linear layer.

        Args:
            module: The torch module to convert. All layers that are
                defined in the ``conversion_map``.
            rpu_config: RPU config to apply to all converted tiles.
                Applied to all converted tiles.
            realistic_read_write: Whether to use closed-loop programming
                when setting the weights. Applied to all converted tiles.
            weight_scaling_omega: If non-zero, applied weights of analog
                layers will be scaled by ``weight_scaling_omega`` divided by
                the absolute maximum value of the original weight matrix.

                Note:
                    Make sure that the weight max and min setting of the
                    device support the desired analog weight range.

        Returns:
            an AnalogLinear layer based on the digital Linear ``module``.
        """
        analog_module = cls(module.in_features,
                            module.out_features,
                            module.bias is not None,
                            rpu_config,
                            realistic_read_write,
                            weight_scaling_omega)

        analog_module.set_weights(module.weight, module.bias)
        return analog_module

    def reset_parameters(self) -> None:
        """Reset the parameters (weight and bias)."""
        super().reset_parameters()
        self.set_weights(self.weight, self.bias)

    def forward(self, x_input: Tensor) -> Tensor:
        """Compute the forward pass."""
        # pylint: disable=arguments-differ
        return AnalogFunction.apply(
            self.analog_tile.get_analog_ctx(), x_input,
            self.analog_tile.shared_weights, not self.training)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Parameter context for analog tiles."""

from typing import Optional, Type, Union, Any, TYPE_CHECKING

from torch import ones, dtype, Tensor
from torch.nn import Parameter
from torch import device as torch_device

if TYPE_CHECKING:
    from aihwkit.simulator.tiles.base import BaseTile


class AnalogContext(Parameter):
    """Context for analog optimizer."""

    def __new__(cls: Type['AnalogContext'], analog_tile: 'BaseTile') -> 'AnalogContext':
        # pylint: disable=signature-differs
        return Parameter.__new__(cls, data=ones((), device=analog_tile.device),
                                 requires_grad=True)

    def __init__(self, analog_tile: 'BaseTile'):
        super().__init__()
        self.analog_tile = analog_tile
        self.use_torch_update = False
        self.use_indexed = False
        self.analog_input = []  # type: list
        self.analog_grad_output = []  # type: list

    def set_data(self, data: Tensor) -> None:
        """Set the data value of the Tensor."""
        # pylint: disable=attribute-defined-outside-init
        self.data.copy_(data)

    def get_data(self) -> Tensor:
        """Get the data value of the underlying Tensor."""
        # pylint: disable=attribute-defined-outside-init
        return self.data.detach()

    def reset(self, analog_tile: Optional['BaseTile'] = None) -> None:
        """Reset the gradient trace and optionally sets the tile pointer."""
        if analog_tile is not None:
            self.analog_tile = analog_tile
            self.analog_tile.analog_ctx = self

        self.analog_input = []
        self.analog_grad_output = []

    def has_gradient(self) -> bool:
        """Return whether a gradient trace was stored."""
        return len(self.analog_input) > 0

    def cuda(
            self,
            device: Optional[Union[torch_device, str, int]] = None
    ) -> 'AnalogContext':
        """Move the context to a cuda device.

        Args:
             device: the desired device of the tile.

        Returns:
            This context in the specified device.
        """
        # pylint: disable=attribute-defined-outside-init
        self.data = self.data.cuda(device)  # type: Tensor

        if not self.analog_tile.is_cuda:
            self.analog_tile = self.analog_tile.cuda(device)
        self.reset(self.analog_tile)

        return self

    def cpu(self) -> 'AnalogContext':
        """Move the context to CPU.

        Note:
            This is a no-op for CPU context.

        Returns:
            self
        """
        self.analog_tile.cpu()  # will raise an error if not possile
        return self

    def to(self, *args: Any, **kwargs: Any) -> 'AnalogContext':
        """Move analog tiles of the current context to a device.

        Note:
            Please be aware that moving analog tiles from GPU to CPU is
            currently not supported.

        Caution:
            Other tensor conversions than moving the device to CUDA,
            such as changing the data type are not supported for analog
            tiles and will be simply ignored.

        Returns:
            This module in the specified device.
        """
        # pylint: disable=invalid-name, attribute-defined-outside-init
        self.data = self.data.to(*args, **kwargs)

        device = None
        if 'device' in kwargs:
            device = kwargs['device']
        elif len(args) > 0 and not isinstance(args[0], (Tensor, dtype)):
            device = torch_device(args[0])

        if device is not None:
            device = torch_device(device)
            if device.type == 'cuda' and not self.analog_tile.is_cuda:
                self.analog_tile = self.analog_tile.cuda(device)
            self.reset(self.analog_tile)

        return self

    def __repr__(self) -> str:
        return 'AnalogContext of ' + self.analog_tile.get_brief_info()



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Utils for configurations presets for resistive processing units."""

# pylint: disable=too-many-instance-attributes

from dataclasses import dataclass

from aihwkit.simulator.configs.utils import (
    BoundManagementType, IOParameters, NoiseManagementType, PulseType,
    UpdateParameters, WeightNoiseType
)


@dataclass
class PresetIOParameters(IOParameters):
    r"""Preset for the forward and backward pass parameters.

    This defines the default (peripheral) hardware configurations
    used for most presets.

    Currently, we assume 7 bit DAC (without stochastic rounding) and 9
    bit ADC (including the sign) and a fixed dynamic range ratio. The
    dynamic range ratio is defined by how many fully-on inputs would
    saturate the output ADC when all weights are set maximal
    conductances. This value is set to 20 here, as the weight range is
    normalized to :math:`-1,\ldots 1`.

    Moreover, the output noise (additive Gaussian) is set to 0.1,
    which is on the order of 1 LSB of the ADC.

    By default, we turned additional weight noise off, however, some
    presets might turn it on as required by the device specification.

    Finally, we assume by default that the device is run with bound
    management (see
    :class:`~aihwkit.simulator.config.utils.BoundManagementType`) and
    noise management (see
    :class:`~aihwkit.simulator.config.utils.NoiseManagementType`)
    turned on to `ITERATIVE` and `ABS_MAX`, respectively.
    """

    bound_management: BoundManagementType = BoundManagementType.ITERATIVE
    noise_management: NoiseManagementType = NoiseManagementType.ABS_MAX

    inp_res: float = 1.0 / (2**7 - 2)  # 7 bit DAC.
    inp_sto_round: bool = False

    out_bound: float = 20.0
    out_noise: float = 0.1
    out_res: float = 1.0 / (2**9 - 2)  # 9 bit ADC.

    # No read noise by default.
    w_noise: float = 0.0
    w_noise_type: WeightNoiseType = WeightNoiseType.NONE


@dataclass
class PresetUpdateParameters(UpdateParameters):
    """Preset for the general update behavior.

    This defines the default update configurations used for most
    presets. Presets might override this default behavior to implement
    other analog SGD optimizers.

    Parallel analog update is the default. We assume stochastic pulse
    to do the parallel update in analog, as described in `Gokmen &
    Vlasov, Front. Neurosci. 2016`_.

    Moreover, we assume that the pulse length is dynamically adjusted
    with a maximal pulse length of 31 pulses.

    .. _`Gokmen & Vlasov, Front. Neurosci. 2016`: \
        https://www.frontiersin.org/articles/10.3389/fnins.2016.00333/full
    """

    desired_bl: int = 31  # Less than 32 preferable (faster implementation).
    pulse_type: PulseType = PulseType.STOCHASTIC_COMPRESSED
    update_bl_management: bool = True  # Dynamically adjusts pulse train length (max 31).
    update_management: bool = True



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: output_file.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import onnx_common_pb2 as onnx__common__pb2
from . import common_pb2 as common__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='output_file.proto',
  package='aihwx',
  syntax='proto2',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x11output_file.proto\x12\x05\x61ihwx\x1a\x11onnx_common.proto\x1a\x0c\x63ommon.proto\"I\n\x10\x45pochInformation\x12\r\n\x05\x65poch\x18\x01 \x02(\x05\x12&\n\x07metrics\x18\x02 \x03(\x0b\x32\x15.aihwx.AttributeProto\"\xa3\x01\n\x0eTrainingOutput\x12\x1f\n\x07version\x18\x01 \x02(\x0b\x32\x0e.aihwx.Version\x12\x1f\n\x07network\x18\x02 \x02(\x0b\x32\x0e.aihwx.Network\x12\'\n\x06\x65pochs\x18\x03 \x03(\x0b\x32\x17.aihwx.EpochInformation\x12&\n\x07metrics\x18\x04 \x03(\x0b\x32\x15.aihwx.AttributeProto'
  ,
  dependencies=[onnx__common__pb2.DESCRIPTOR,common__pb2.DESCRIPTOR,])




_EPOCHINFORMATION = _descriptor.Descriptor(
  name='EpochInformation',
  full_name='aihwx.EpochInformation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='epoch', full_name='aihwx.EpochInformation.epoch', index=0,
      number=1, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metrics', full_name='aihwx.EpochInformation.metrics', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=61,
  serialized_end=134,
)


_TRAININGOUTPUT = _descriptor.Descriptor(
  name='TrainingOutput',
  full_name='aihwx.TrainingOutput',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='aihwx.TrainingOutput.version', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='network', full_name='aihwx.TrainingOutput.network', index=1,
      number=2, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='epochs', full_name='aihwx.TrainingOutput.epochs', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metrics', full_name='aihwx.TrainingOutput.metrics', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=137,
  serialized_end=300,
)

_EPOCHINFORMATION.fields_by_name['metrics'].message_type = onnx__common__pb2._ATTRIBUTEPROTO
_TRAININGOUTPUT.fields_by_name['version'].message_type = common__pb2._VERSION
_TRAININGOUTPUT.fields_by_name['network'].message_type = common__pb2._NETWORK
_TRAININGOUTPUT.fields_by_name['epochs'].message_type = _EPOCHINFORMATION
_TRAININGOUTPUT.fields_by_name['metrics'].message_type = onnx__common__pb2._ATTRIBUTEPROTO
DESCRIPTOR.message_types_by_name['EpochInformation'] = _EPOCHINFORMATION
DESCRIPTOR.message_types_by_name['TrainingOutput'] = _TRAININGOUTPUT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

EpochInformation = _reflection.GeneratedProtocolMessageType('EpochInformation', (_message.Message,), {
  'DESCRIPTOR' : _EPOCHINFORMATION,
  '__module__' : 'output_file_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.EpochInformation)
  })
_sym_db.RegisterMessage(EpochInformation)

TrainingOutput = _reflection.GeneratedProtocolMessageType('TrainingOutput', (_message.Message,), {
  'DESCRIPTOR' : _TRAININGOUTPUT,
  '__module__' : 'output_file_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.TrainingOutput)
  })
_sym_db.RegisterMessage(TrainingOutput)


# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Mappings for version 1 of the AIHW Composer format."""

from collections import namedtuple
from typing import Any, Dict

from torch.nn import (
    BCELoss, BatchNorm2d, Conv2d, ConvTranspose2d, CrossEntropyLoss, Flatten,
    LeakyReLU, Linear, LogSigmoid, LogSoftmax, MSELoss, MaxPool2d, NLLLoss,
    ReLU, Sigmoid, Softmax, Tanh
)
from torchvision.datasets import FashionMNIST, SVHN

from aihwkit.cloud.converter.definitions.onnx_common_pb2 import AttributeProto
from aihwkit.cloud.converter.exceptions import ConversionError
from aihwkit.nn import AnalogConv2d, AnalogLinear
from aihwkit.optim import AnalogSGD
from aihwkit.simulator.presets import (
    CapacitorPreset, EcRamPreset, IdealizedPreset, ReRamESPreset, ReRamSBPreset,
    PCMPreset,
    MixedPrecisionCapacitorPreset, MixedPrecisionEcRamPreset,
    MixedPrecisionIdealizedPreset, MixedPrecisionPCMPreset,
    MixedPrecisionReRamESPreset, MixedPrecisionReRamSBPreset,
    TikiTakaCapacitorPreset, TikiTakaEcRamPreset, TikiTakaIdealizedPreset,
    TikiTakaReRamESPreset, TikiTakaReRamSBPreset
)

Type = namedtuple('Type', ['attribute_type', 'field', 'fn'])

TYPES = {
    int: Type(AttributeProto.AttributeType.INT, 'i', lambda x: x),  # type: ignore
    bool: Type(AttributeProto.AttributeType.BOOL, 'b', lambda x: x),  # type: ignore
    str: Type(AttributeProto.AttributeType.STRING, 's',  # type: ignore
              lambda x: x.encode('utf-8')),
    float: Type(AttributeProto.AttributeType.FLOAT, 'f', lambda x: x)  # type: ignore
}

TYPES_LISTS = {
    int: Type(AttributeProto.AttributeType.INTS, 'ints', lambda x: x),  # type: ignore
    bool: Type(AttributeProto.AttributeType.BOOLS, 'bools', lambda x: x),  # type: ignore
    str: Type(AttributeProto.AttributeType.STRINGS, 'strings',  # type: ignore
              lambda x: [y.encode('utf-8') for y in x]),
    float: Type(AttributeProto.AttributeType.FLOATS, 'floats', lambda x: x)  # type: ignore
}


class Function:
    """Mapping for a function-like entity."""

    def __init__(
            self,
            id_: str,
            args: Dict
    ):
        self.id_ = id_
        self.args = args

    def to_proto(self, source: object, proto_cls: type) -> object:
        """Convert a source object into a destination object."""
        instance = proto_cls(id=self.id_)

        for name, type_ in self.args.items():
            value = self.get_field_value_to_proto(source, name, None)

            argument = AttributeProto(name=name)
            if isinstance(type_, list):
                proto_type = TYPES_LISTS[type_[0]]
                final_value = proto_type.fn(value)

                # Fix arguments where ``Union[T, Sequence[T]]`` is allowed.
                # TODO: generalize, as this check assumes int and 2-size.
                if isinstance(final_value, int):
                    final_value = [final_value, final_value]

                getattr(argument, proto_type.field).extend(final_value)
            else:
                proto_type = TYPES[type_]
                setattr(argument, proto_type.field, proto_type.fn(value))
            argument.type = proto_type.attribute_type
            instance.arguments.append(argument)

        return instance

    def from_proto(self, source: Any, cls: type) -> object:
        """Convert a proto object into a destination object."""
        kwargs = {}

        for argument in source.arguments:
            type_ = self.args[argument.name]

            if isinstance(type_, list):
                proto_type = TYPES_LISTS[type_[0]]
            else:
                proto_type = TYPES[type_]

            new_argument = self.get_argument_from_proto(argument, proto_type.field, None)
            if isinstance(type_, list):
                new_argument[argument.name] = list(new_argument[argument.name])
            kwargs.update(new_argument)

        return cls(**kwargs)

    def get_field_value_to_proto(self, source: Any, field: str, default: Any = None) -> Any:
        """Get the value of a field."""
        return getattr(source, field, default)

    def get_argument_from_proto(self, source: Any, field: str, default: Any = None) -> Dict:
        """Get the value of an argument."""
        return {source.name: getattr(source, field, default)}


class LayerFunction(Function):
    """Mapping for a function-like entity (Layer)."""

    def get_field_value_to_proto(self, source: Any, field: str, default: Any = None) -> Any:
        """Get the value of a field."""
        if field == 'bias':
            return getattr(source, 'bias', None) is not None
        if field == 'rpu_config':
            preset_cls = type(source.analog_tile.rpu_config)
            try:
                return Mappings.presets[preset_cls]
            except KeyError as ex:
                raise ConversionError('Invalid rpu_config in layer: {} not '
                                      'among the presets'.format(preset_cls)) from ex

        return super().get_field_value_to_proto(source, field, default)

    def get_argument_from_proto(self, source: Any, field: str, default: Any = None) -> Dict:
        """Get the value of an argument."""
        if source.name == 'rpu_config':
            preset_str = getattr(source, field, default).decode('utf-8')
            try:
                preset = InverseMappings.presets[preset_str]
            except KeyError as ex:
                raise ConversionError('Invalid rpu_config in layer: {} not '
                                      'among the presets'.format(preset_str)) from ex
            return {'rpu_config': preset()}

        return super().get_argument_from_proto(source, field, default)


class Mappings:
    """Mappings between Python entities and AIHW format."""
    # pylint: disable=too-few-public-methods

    datasets = {
        FashionMNIST: 'fashion_mnist',
        SVHN: 'svhn'
    }

    layers = {
        AnalogConv2d: LayerFunction('AnalogConv2d', {
            'in_channels': int,
            'out_channels': int,
            'kernel_size': [int],
            'stride': [int],
            'padding': [int],
            'dilation': [int],
            'bias': bool,
            'rpu_config': str,
            'weight_scaling_omega': float
        }),
        AnalogLinear: LayerFunction('AnalogLinear', {
            'in_features': int,
            'out_features': int,
            'bias': bool,
            'rpu_config': str,
            'weight_scaling_omega': float
        }),
        BatchNorm2d: LayerFunction('BatchNorm2d', {
            'num_features': int
        }),
        Conv2d: LayerFunction('Conv2d', {
            'in_channels': int,
            'out_channels': int,
            'kernel_size': [int],
            'stride': [int],
            'padding': [int],
            'dilation': [int],
            'bias': bool
        }),
        ConvTranspose2d: LayerFunction('ConvTranspose2d', {
            'in_channels': int,
            'out_channels': int,
            'kernel_size': [int],
            'stride': [int],
            'padding': [int],
            'output_padding': [int],
            'dilation': [int],
            'bias': bool
        }),
        Flatten: LayerFunction('Flatten', {}),
        Linear: LayerFunction('Linear', {
            'in_features': int,
            'out_features': int,
            'bias': bool
        }),
        MaxPool2d: LayerFunction('MaxPool2d', {
            'kernel_size': int,
            'stride': int,
            'padding': int,
            'dilation': int,
            'ceil_mode': bool
        })
    }

    activation_functions = {
        LeakyReLU: Function('LeakyReLU', {
            'negative_slope': float
        }),
        LogSigmoid: Function('LogSigmoid', {}),
        LogSoftmax: Function('LogSoftmax', {
            'dim': int
        }),
        ReLU: Function('ReLU', {}),
        Sigmoid: Function('Sigmoid', {}),
        Softmax: Function('Softmax', {
            'dim': int
        }),
        Tanh: Function('Tanh', {})
    }

    loss_functions = {
        BCELoss: Function('BCELoss', {}),
        CrossEntropyLoss: Function('CrossEntropyLoss', {}),
        MSELoss: Function('MSELoss', {}),
        NLLLoss: Function('NLLLoss', {})
    }

    optimizers = {
        AnalogSGD: Function('AnalogSGD', {
            'lr': float
        })
    }

    presets = {
        ReRamESPreset: 'ReRamESPreset',
        ReRamSBPreset: 'ReRamSBPreset',
        CapacitorPreset: 'CapacitorPreset',
        EcRamPreset: 'EcRamPreset',
        IdealizedPreset: 'IdealizedPreset',
        PCMPreset: 'PCMPreset',
        MixedPrecisionReRamESPreset: 'MixedPrecisionReRamESPreset',
        MixedPrecisionReRamSBPreset: 'MixedPrecisionReRamSBPreset',
        MixedPrecisionCapacitorPreset: 'MixedPrecisionCapacitorPreset',
        MixedPrecisionEcRamPreset: 'MixedPrecisionEcRamPreset',
        MixedPrecisionIdealizedPreset: 'MixedPrecisionIdealizedPreset',
        MixedPrecisionPCMPreset: 'MixedPrecisionPCMPreset',
        TikiTakaReRamESPreset: 'TikiTakaReRamESPreset',
        TikiTakaReRamSBPreset: 'TikiTakaReRamSBPreset',
        TikiTakaCapacitorPreset: 'TikiTakaCapacitorPreset',
        TikiTakaEcRamPreset: 'TikiTakaEcRamPreset',
        TikiTakaIdealizedPreset: 'TikiTakaIdealizedPreset',
    }


def build_inverse_mapping(mapping: Dict) -> Dict:
    """Create the inverse mapping between Python entities and AIHW Composer formats."""
    return {value if not isinstance(value, Function) else value.id_: key
            for key, value in mapping.items()}


class InverseMappings:
    """Mappings between AIHW Composer format and Python entities."""
    # pylint: disable=too-few-public-methods

    datasets = build_inverse_mapping(Mappings.datasets)
    layers = build_inverse_mapping(Mappings.layers)
    activation_functions = build_inverse_mapping(Mappings.activation_functions)
    loss_functions = build_inverse_mapping(Mappings.loss_functions)
    optimizers = build_inverse_mapping(Mappings.optimizers)
    presets = build_inverse_mapping(Mappings.presets)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Visualization utilities.

This module includes plotting and visualization utilities for ``aihwkit``.
Using this module has extra dependencies that can be installed via the
extras mechanism::

    pip install aihwkit[visualization]
"""

from typing import Tuple, Any, Union
from copy import deepcopy

from matplotlib.figure import Figure
from numpy import ndarray

from torch import ones, eye, from_numpy
from torch import device as torch_device

import numpy as np
import matplotlib.ticker as ticker
import matplotlib.pyplot as plt

from aihwkit.simulator.configs import SingleRPUConfig, UnitCellRPUConfig
from aihwkit.exceptions import ConfigError
from aihwkit.simulator.tiles import (
    AnalogTile, BaseTile
)
from aihwkit.simulator.rpu_base import cuda
from aihwkit.simulator.configs.devices import PulsedDevice, UnitCell
from aihwkit.simulator.configs.utils import (
    IOParameters, WeightNoiseType, NoiseManagementType, BoundManagementType,
    PulseType, UpdateParameters
)


def compute_pulse_response(
        analog_tile: BaseTile,
        direction: ndarray,
        use_forward: bool = False
) -> ndarray:
    """Compute the pulse response of a given device configuration.

    Args:
        analog_tile: Base tile used for computing the weight traces
        direction: numpy vector of directions to sequentially apply (-1 or 1)
        use_forward: Whether to use the (noisy) forward pass to read out the weights
            (otherwise returns exact weight value).

    Returns:
        An numpy array ``w_trace`` of dimensions ``len(direction) x out_size x in_size``
    """
    out_size = analog_tile.out_size
    in_size = analog_tile.in_size

    if analog_tile.is_cuda:
        device = torch_device('cuda')
    else:
        device = torch_device('cpu')

    total_iters = len(direction)
    w_trace = np.zeros((total_iters, out_size, in_size))
    in_vector = -ones(1, in_size, device=device)
    out_vector = ones(1, out_size, device=device)
    in_eye = eye(in_size, device=device)
    dir_tensor = from_numpy(direction).float().to(device)

    for i in range(total_iters):
        # Update the pulses.
        analog_tile.update(in_vector, out_vector * dir_tensor[i])

        if use_forward:
            # Save weights by using the forward pass (to get the short-term read noise).
            w_trace[i, :, :] = analog_tile.forward(in_eye).detach().cpu().numpy().T
        else:
            # Noise free.
            w_trace[i, :, :] = analog_tile.get_weights()[0].detach().cpu().numpy()

    return w_trace


def plot_pulse_response(
        analog_tile: BaseTile,
        direction: ndarray,
        use_forward: bool = False
) -> ndarray:
    """Plot the pulse response of a direction vector for each weight of
    the analog tile.

    Args:
        analog_tile: Base tile used for computing the weight traces
        direction: vector of directions to sequentially apply (-1 or 1)
        use_forward: Whether to use the (noisy) forward pass to read out the weights
            (otherwise returns exact weight value).

    Returns:
        w_trace: from :func:`compute_pulse_response`
    """
    w_trace = compute_pulse_response(analog_tile, direction, use_forward)

    plt.plot(w_trace.reshape(w_trace.shape[0], -1))
    if use_forward:
        plt.title(analog_tile.rpu_config.device.__class__.__name__)
    else:
        plt.title('{} (without cycle/read noise)'
                  .format(analog_tile.rpu_config.device.__class__.__name__))
    plt.ylabel('Weight [conductance]')
    plt.xlabel('Pulse number #')

    return w_trace


def compute_pulse_statistics(
        w_nodes: ndarray,
        w_trace: ndarray,
        direction: ndarray,
        up_direction: bool,
        smoothness: float = 0.5
) -> Tuple[ndarray, ndarray]:
    """Compute the statistics of the step trace from :func:`compute_pulse_response`.

    Args:
        w_nodes: weight range vector to estimate the step histogram
        w_trace: weight trace from :func:`compute_pulse_response`
        direction: direction vector used to generate the weight traces
        up_direction: whether and plot to compute the statistics for up or down direction
        smoothness: value for smoothing the estimation of the
            statistical step response curves

    Returns:
        Tuple of ``(dw_mean, dw_std)``.
    """
    # pylint: disable=too-many-locals

    def calc_mean_and_std(
            node: ndarray,
            w_values: ndarray,
            delta_w: ndarray,
            lam: float
    ) -> Tuple[ndarray, ndarray]:
        """Calculate the mean and std of a w location (node).

        Note:
            In case there are multiple trials then it also includes
            device-to-device variation.
        """
        alpha = np.exp(-0.5*(node - w_values)**2/lam**2)
        beta = alpha.sum(axis=0)
        alpha[:, beta < 0.1] = np.nan
        alpha /= np.expand_dims(beta, axis=0)
        mean = np.sum(alpha*delta_w, axis=0)
        std = np.sqrt(np.sum(alpha*(delta_w - np.expand_dims(mean, axis=0))**2, axis=0))
        return (mean, std)

    # dw statistics.
    delta_w = np.diff(w_trace, axis=0)
    w_trace_s = w_trace[:-1, :, :]
    if up_direction:
        msk = np.logical_and(direction[:-1] > 0, np.diff(direction) == 0)
        w_values = w_trace_s[msk, :, :]
        delta_w_values = delta_w[msk, :, :]
    else:
        msk = np.logical_and(direction[:-1] < 0, np.diff(direction) == 0)
        w_values = w_trace_s[msk, :, :]
        delta_w_values = delta_w[msk, :, :]

    dw_mean = np.zeros((len(w_nodes), w_trace.shape[1], w_trace.shape[2]))
    dw_std = np.zeros((len(w_nodes), w_trace.shape[1], w_trace.shape[2]))

    lam = (w_nodes[1]-w_nodes[0])/2*smoothness
    for i, node in enumerate(w_nodes):
        dw_mean[i, :, :], dw_std[i, :, :] = calc_mean_and_std(node, w_values, delta_w_values, lam)

    return dw_mean, dw_std


def plot_pulse_statistics(
        w_trace: ndarray,
        direction: ndarray,
        up_direction: bool,
        num_nodes: int = 100,
        smoothness: float = 0.5
) -> Tuple[ndarray, ndarray, ndarray]:
    """Plot the dG-G curve from a given weight trace and direction vector.

    Args:
        w_trace: weight trace from :func:`compute_pulse_response`
        direction: direction vector used to generate ``w_trace``
        up_direction: whether and plot to compute the statistics for up or down direction
        num_nodes: number of nodes for estimation of the step histogram
        smoothness: value for smoothing the estimation of the
            statistical step response curves

    Returns:
        A tuple (w_nodes, dw_mean, dw_std) from :func:`compute_pulse_statistics`
    """

    def errorbar_patch(x: ndarray, mean: ndarray, std: ndarray) -> None:
        """Plot a patchy error bar."""
        axis = plt.plot(x, mean)[0]
        plt.fill_between(x, mean - std, mean + std, edgecolor=None,
                         facecolor=axis.get_color(), alpha=0.5)

    # Compute statistics.
    w_nodes = np.linspace(w_trace.min(), w_trace.max(), num_nodes)
    dw_mean, dw_std = compute_pulse_statistics(w_nodes, w_trace, direction, up_direction,
                                               smoothness)

    n_traces = dw_mean.shape[1] * dw_mean.shape[2]

    for i in range(n_traces):
        errorbar_patch(w_nodes,
                       dw_mean.reshape(-1, n_traces)[:, i],
                       dw_std.reshape(-1, n_traces)[:, i])

    plt.xlabel('Weight $w$')
    plt.ylabel('Avg. step $\\Delta w$')
    if up_direction:
        plt.title('up-direction')
    else:
        plt.title('down-direction')

    return w_nodes, dw_mean, dw_std


def get_tile_for_plotting(
        rpu_config: Union[SingleRPUConfig, UnitCellRPUConfig],
        n_traces: int,
        use_cuda: bool = False,
        noise_free: bool = False
) -> BaseTile:
    """Return an analog tile for plotting the response curve.

    Args:
        rpu_config: RPU Configuration to use for plotting
        n_traces: Number of traces to plot
        use_cuda: Whether to use the CUDA implementation (if available)
        noise_free: Whether to turn-off cycle-to-cycle noises (if possible)

    Returns:
        Instantiated tile.
    """

    def set_noise_free(dev: Any) -> Any:
        if hasattr(dev, 'dw_min_std'):
            dev.dw_min_std = 0.0  # Noise free.

        if hasattr(dev, 'refresh_forward'):
            setattr(dev, 'refresh_forward', IOParameters(is_perfect=True))

        if hasattr(dev, 'refresh_update'):
            setattr(dev, 'refresh_update', UpdateParameters(pulse_type=PulseType.NONE))

        if hasattr(dev, 'transfer_forward'):
            setattr(dev, 'refresh_forward', IOParameters(is_perfect=True))

        if hasattr(dev, 'transfer_update'):
            setattr(dev, 'transfer_update', UpdateParameters(pulse_type=PulseType.NONE))

        if (hasattr(dev, 'write_noise_std') and
           getattr(dev, 'write_noise_std') > 0.0):
            # Just make very small to avoid hidden parameter mismatch.
            setattr(dev, 'write_noise_std', 1e-6)

    config = deepcopy(rpu_config)

    # Make sure we use single pulses for the overview.
    config.update.update_bl_management = False
    config.update.update_management = False
    config.update.desired_bl = 1

    if noise_free:
        config.forward.is_perfect = True

        set_noise_free(config.device)
        if hasattr(config.device, 'unit_cell_devices'):
            for dev in getattr(config.device, 'unit_cell_devices'):
                set_noise_free(dev)
        if hasattr(config.device, 'device'):
            set_noise_free(getattr(config.device, 'device'))

    analog_tile = AnalogTile(n_traces, 1, config)  # type: BaseTile
    analog_tile.set_learning_rate(1)
    w_min = getattr(config.device.as_bindings(), 'w_min', -1.0)

    weights = w_min * ones((n_traces, 1))
    analog_tile.set_weights(weights)

    if use_cuda and cuda.is_compiled():
        return analog_tile.cuda()
    return analog_tile


def estimate_n_steps(rpu_config: Union[SingleRPUConfig, UnitCellRPUConfig]) -> int:
    """Estimate the n_steps.

    Note:
        The estimate of the number of update pulses needed to drive
        from smallest to largest conductance. The estimation just
        assumes linear behavior, thus only be a rough estimate for
        non-linear response curves.

    Args:
        rpu_config: RPU Configuration to use for plotting

    Returns:
        Guessed number of steps

    Raises:
        ConfigError: If rpu_config.device does not have the w_min
            attribute (which is only ensured for
            :class:`~aihwkit.simulator.configs.devices.PulseDevice`)
    """
    if not isinstance(rpu_config, SingleRPUConfig):
        return 1000

    device_binding = rpu_config.device.as_bindings()

    if not hasattr(device_binding, 'w_min'):
        raise ConfigError('n_step estimation only for PulsedDevice. ' +
                          'Provide n_step explicitly.')

    weight_granularity = device_binding.calc_weight_granularity()
    w_min = device_binding.w_min
    w_max = device_binding.w_max

    n_steps = int(np.round((w_max - w_min) / weight_granularity))
    return n_steps


def plot_response_overview(
        rpu_config: Union[SingleRPUConfig, UnitCellRPUConfig],
        n_loops: int = 5,
        n_steps: int = None,
        n_traces: int = 5,
        use_cuda: bool = False,
        smoothness: float = 0.1
) -> None:
    """Plot the step response and statistics of a given device configuration.

    Args:
        rpu_config: RPU Configuration to use for plotting
        n_loops: How many hyper-cycles (up/down pulse sequences) to plot
        n_steps: Number of up/down steps per cycle. If not given, will
            be tried to be estimated (only for ``PulsedDevice``
            possible otherwise defaults to 1000 if ``n_steps=None``).
        n_traces: Number of traces to plot
        use_cuda: Whether to use the CUDA implementation (if available)
        smoothness: value for smoothing the estimation of the
            statistical step response curves
    """
    if n_steps is None:
        n_steps = estimate_n_steps(rpu_config)

    total_iters = min(max(n_loops*2*n_steps, 1000), max(50000, 2*n_steps))
    direction = np.sign(np.sin(np.pi*(np.arange(total_iters)+1)/n_steps))

    plt.clf()

    # 1. Noisy tile.
    analog_tile = get_tile_for_plotting(rpu_config, n_traces, use_cuda, noise_free=False)
    plt.subplot(3, 1, 1)
    plot_pulse_response(analog_tile, direction, use_forward=True)

    # 2. Noise-free tile.
    analog_tile_noise_free = get_tile_for_plotting(rpu_config, n_traces, use_cuda, noise_free=True)
    analog_tile_noise_free.set_hidden_parameters(analog_tile.get_hidden_parameters())

    plt.subplot(3, 1, 2)
    w_trace = plot_pulse_response(analog_tile_noise_free, direction, use_forward=False)

    num_nodes = min(n_steps, 100)
    # 3. Plot up statistics.
    plt.subplot(3, 2, 5)
    plot_pulse_statistics(w_trace, direction, True, num_nodes, smoothness)

    # 4. Plot down statistics.
    plt.subplot(3, 2, 6)
    plot_pulse_statistics(w_trace, direction, False, num_nodes, smoothness)

    plt.tight_layout()


def plot_device(device: Union[PulsedDevice, UnitCell], w_noise: float = 0.0, **kwargs: Any) -> None:
    """Plot the step response figure for a given device (preset).

    Note:
        It will use an amount of read weight noise ``w_noise`` for
        reading the weights.

    Args:
        device: PulsedDevice parameters
        w_noise: Weight noise standard deviation during read
        kwargs: for other parameters, see :func:`plot_response_overview`
    """
    plt.figure(figsize=[7, 7])
    # To simulate some weight read noise.
    io_pars = IOParameters(out_noise=0.0,    # no out noise
                           w_noise=w_noise,  # quite low
                           inp_res=-1.,      # turn off DAC
                           out_bound=100.,   # not limiting
                           out_res=-1.,      # turn off ADC
                           bound_management=BoundManagementType.NONE,
                           noise_management=NoiseManagementType.NONE,
                           w_noise_type=WeightNoiseType.ADDITIVE_CONSTANT)

    if isinstance(device, PulsedDevice):
        plot_response_overview(SingleRPUConfig(device=device, forward=io_pars), **kwargs)
    else:
        plot_response_overview(UnitCellRPUConfig(device=device, forward=io_pars), **kwargs)


def plot_device_compact(
        device: Union[PulsedDevice, UnitCell],
        w_noise: float = 0.0,
        n_steps: int = None,
        n_traces: int = 3,
        use_cuda: bool = False,
) -> Figure:
    """Plot a compact step response figure for a given device (preset).

    Note:
        It will use an amount of read weight noise ``w_noise`` for
        reading the weights.

    Args:
        device: ``PulsedDevice`` or ``UnitCell`` parameters
        w_noise: Weight noise standard deviation during read
        n_steps: Number of up/down steps per cycle. If not given, will
            be tried to be estimated (only for ``PulsedDevice``
            possible otherwise defaults to 1000 if ``n_steps=None``).
        n_traces: Number of traces to plot (for device-to-device variation)
        use_cuda: Whether to use CUDA for the computation

    Returns:
        the compact step response figure.
    """
    # pylint: disable=too-many-locals,too-many-statements
    def get_rpu_config(device: Union[PulsedDevice, UnitCell], io_pars: IOParameters) \
            -> Union[SingleRPUConfig, UnitCellRPUConfig]:
        if isinstance(device, PulsedDevice):
            return SingleRPUConfig(device=device, forward=io_pars)
        return UnitCellRPUConfig(device=device, forward=io_pars)

    figure = plt.figure(figsize=[12, 4])

    # To simulate some weight read noise.
    io_pars = IOParameters(out_noise=0.0,    # no out noise
                           w_noise=w_noise,  # quite low
                           inp_res=-1.,      # turn off DAC
                           out_bound=100.,   # not limiting
                           out_res=-1.,      # turn off ADC
                           bound_management=BoundManagementType.NONE,
                           noise_management=NoiseManagementType.NONE,
                           w_noise_type=WeightNoiseType.ADDITIVE_CONSTANT)

    rpu_config = get_rpu_config(device, io_pars)

    if n_steps is None:
        n_steps = estimate_n_steps(rpu_config)

    # Noisy tile response curves.
    n_loops = 2
    total_iters = n_loops*2*n_steps
    direction = np.sign(np.sin(np.pi*(np.arange(total_iters)+1)/n_steps))

    analog_tile = get_tile_for_plotting(rpu_config, n_traces, use_cuda, noise_free=False)
    w_trace = compute_pulse_response(analog_tile, direction, use_forward=True)\
        .reshape(-1, n_traces)
    axis = figure.add_subplot(1, 1, 1)
    axis.plot(w_trace, linewidth=1)
    axis.set_title(analog_tile.rpu_config.device.__class__.__name__)
    axis.set_xlabel('Pulse number #')
    limit = np.abs(w_trace).max()*1.2
    axis.set_ylim(-limit, limit)
    axis.set_xlim(0, total_iters-1)
    axis.xaxis.set_major_formatter(ticker.ScalarFormatter(useMathText=True))

    # Noise-free tile for statistics.
    n_loops = 1
    total_iters = min(max(n_loops*2*n_steps, 1000), max(50000, 2*n_steps))
    direction = np.sign(np.sin(np.pi*(np.arange(total_iters)+1)/n_steps))

    analog_tile_noise_free = get_tile_for_plotting(rpu_config, n_traces, use_cuda, noise_free=True)
    analog_tile_noise_free.set_hidden_parameters(analog_tile.get_hidden_parameters())

    w_trace = compute_pulse_response(analog_tile_noise_free, direction, False)

    # Compute statistics.
    num_nodes = min(n_steps, 100)
    w_nodes = np.linspace(w_trace.min(), w_trace.max(), num_nodes)

    dw_mean_up = compute_pulse_statistics(w_nodes, w_trace, direction, True)[0]\
        .reshape(-1, n_traces)
    dw_mean_down = compute_pulse_statistics(w_nodes, w_trace, direction, False)[0]\
        .reshape(-1, n_traces)

    # Plot mean up statistics.
    pos = axis.get_position().bounds
    space = 0.1
    gap = 0.01
    axis.set_position([pos[0] + gap + space, pos[1], pos[2] - 2*gap - 2*space, pos[3]])
    axis.set_yticks([])

    axis_left = figure.add_axes([pos[0], pos[1], space, pos[3]])
    dw_mean_up = dw_mean_up.reshape(-1, n_traces)
    for i in range(n_traces):
        axis_left.plot(dw_mean_up[:, i], w_nodes)

    axis_left.set_position([pos[0], pos[1], space, pos[3]])
    axis_left.set_xlabel('Up pulse size')
    axis_left.set_ylabel('Weight \n [conductance]')
    axis_left.set_ylim(-limit, limit)

    # Plot mean down statistics.
    axis_right = figure.add_axes([pos[0] + pos[2] - space, pos[1], space, pos[3]])
    dw_mean_down = dw_mean_down.reshape(-1, n_traces)
    for i in range(n_traces):
        axis_right.plot(np.abs(dw_mean_down[:, i]), w_nodes)

    axis_right.set_yticks([])
    axis_right.set_xlabel('Down pulse size')
    axis_right.set_ylim(-limit, limit)

    # Set xlim's.
    limit = np.maximum(np.nanmax(np.abs(dw_mean_down)),
                       np.nanmax(np.abs(dw_mean_up))) * 1.2
    axis_left.set_xlim(0.0, limit)
    axis_right.set_xlim(0.0, limit)

    return figure


def plot_device_symmetry(
        device: PulsedDevice,
        w_noise: float = 0.0,
        n_pulses: int = 10000,
        n_traces: int = 3,
        use_cuda: bool = False,
        w_init: float = 1.0,
) -> None:
    """Plot the response figure for a given device (preset).

    It will show the response to alternating up down pulses.

    Note:
        It will use an amount of read weight noise ``w_noise`` for
        reading the weights.

    Args:
        device: PulsedDevice parameters
        n_pulses: total number of pulses
        w_noise: Weight noise standard deviation during read
        n_traces: Number of device traces
        use_cuda: Whether to use CUDA,
        w_init: Initial value of the weights
    """
    plt.figure(figsize=[10, 5])

    io_pars = IOParameters(out_noise=0.0,    # no out noise
                           w_noise=w_noise,  # quite low
                           inp_res=-1.,      # turn off DAC
                           out_bound=100.,   # not limiting
                           out_res=-1.,      # turn off ADC
                           bound_management=BoundManagementType.NONE,
                           noise_management=NoiseManagementType.NONE,
                           w_noise_type=WeightNoiseType.ADDITIVE_CONSTANT)

    rpu_config = SingleRPUConfig(device=device, forward=io_pars)

    direction = np.sign(np.cos(np.pi*np.arange(n_pulses)))
    plt.clf()

    analog_tile = get_tile_for_plotting(rpu_config, n_traces, use_cuda, noise_free=False)
    weights = w_init*ones((n_traces, 1))
    analog_tile.set_weights(weights)

    plot_pulse_response(analog_tile, direction, use_forward=False)
    plt.ylim([-1, 1])
    plt.grid(True)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: input_file.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import common_pb2 as common__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='input_file.proto',
  package='aihwx',
  syntax='proto2',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x10input_file.proto\x12\x05\x61ihwx\x1a\x0c\x63ommon.proto\"\\\n\x07\x44\x61taset\x12\x12\n\ndataset_id\x18\x01 \x02(\t\x12)\n\ntransforms\x18\x02 \x03(\x0b\x32\x15.aihwx.TransformProto\x12\x12\n\nbatch_size\x18\x03 \x01(\x05\"\x9f\x01\n\x08Training\x12(\n\toptimizer\x18\x01 \x02(\x0b\x32\x15.aihwx.OptimizerProto\x12(\n\tscheduler\x18\x02 \x01(\x0b\x32\x15.aihwx.SchedulerProto\x12/\n\rloss_function\x18\x03 \x01(\x0b\x32\x18.aihwx.LossFunctionProto\x12\x0e\n\x06\x65pochs\x18\x04 \x01(\x05\"\x95\x01\n\rTrainingInput\x12\x1f\n\x07version\x18\x01 \x02(\x0b\x32\x0e.aihwx.Version\x12\x1f\n\x07\x64\x61taset\x18\x02 \x02(\x0b\x32\x0e.aihwx.Dataset\x12\x1f\n\x07network\x18\x03 \x02(\x0b\x32\x0e.aihwx.Network\x12!\n\x08training\x18\x04 \x02(\x0b\x32\x0f.aihwx.Training'
  ,
  dependencies=[common__pb2.DESCRIPTOR,])




_DATASET = _descriptor.Descriptor(
  name='Dataset',
  full_name='aihwx.Dataset',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='dataset_id', full_name='aihwx.Dataset.dataset_id', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='transforms', full_name='aihwx.Dataset.transforms', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='batch_size', full_name='aihwx.Dataset.batch_size', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=41,
  serialized_end=133,
)


_TRAINING = _descriptor.Descriptor(
  name='Training',
  full_name='aihwx.Training',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='optimizer', full_name='aihwx.Training.optimizer', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='scheduler', full_name='aihwx.Training.scheduler', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='loss_function', full_name='aihwx.Training.loss_function', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='epochs', full_name='aihwx.Training.epochs', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=136,
  serialized_end=295,
)


_TRAININGINPUT = _descriptor.Descriptor(
  name='TrainingInput',
  full_name='aihwx.TrainingInput',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='aihwx.TrainingInput.version', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dataset', full_name='aihwx.TrainingInput.dataset', index=1,
      number=2, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='network', full_name='aihwx.TrainingInput.network', index=2,
      number=3, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='training', full_name='aihwx.TrainingInput.training', index=3,
      number=4, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=298,
  serialized_end=447,
)

_DATASET.fields_by_name['transforms'].message_type = common__pb2._TRANSFORMPROTO
_TRAINING.fields_by_name['optimizer'].message_type = common__pb2._OPTIMIZERPROTO
_TRAINING.fields_by_name['scheduler'].message_type = common__pb2._SCHEDULERPROTO
_TRAINING.fields_by_name['loss_function'].message_type = common__pb2._LOSSFUNCTIONPROTO
_TRAININGINPUT.fields_by_name['version'].message_type = common__pb2._VERSION
_TRAININGINPUT.fields_by_name['dataset'].message_type = _DATASET
_TRAININGINPUT.fields_by_name['network'].message_type = common__pb2._NETWORK
_TRAININGINPUT.fields_by_name['training'].message_type = _TRAINING
DESCRIPTOR.message_types_by_name['Dataset'] = _DATASET
DESCRIPTOR.message_types_by_name['Training'] = _TRAINING
DESCRIPTOR.message_types_by_name['TrainingInput'] = _TRAININGINPUT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Dataset = _reflection.GeneratedProtocolMessageType('Dataset', (_message.Message,), {
  'DESCRIPTOR' : _DATASET,
  '__module__' : 'input_file_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.Dataset)
  })
_sym_db.RegisterMessage(Dataset)

Training = _reflection.GeneratedProtocolMessageType('Training', (_message.Message,), {
  'DESCRIPTOR' : _TRAINING,
  '__module__' : 'input_file_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.Training)
  })
_sym_db.RegisterMessage(Training)

TrainingInput = _reflection.GeneratedProtocolMessageType('TrainingInput', (_message.Message,), {
  'DESCRIPTOR' : _TRAININGINPUT,
  '__module__' : 'input_file_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.TrainingInput)
  })
_sym_db.RegisterMessage(TrainingInput)


# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Basic training Experiment."""

from typing import Any, Dict, List, Tuple, Type

from torch import device as torch_device, max as torch_max, Tensor
from torch.nn import Module, NLLLoss
from torch.nn.modules.loss import _Loss
from torch.optim import Optimizer
from torch.utils.data import DataLoader, Dataset, Subset
from torchvision.datasets import FashionMNIST, SVHN
from torchvision.transforms import Compose, Normalize, ToTensor

from aihwkit.experiments.experiments.base import Experiment, Signals
from aihwkit.optim import AnalogSGD


class BasicTraining(Experiment):
    """Experiment for training a neural network.

    ``Experiment`` that represents training a neural network using a basic
    training loop.

    This class contains:

    * the data needed for an experiment. The recommended way of setting this
      data is via the arguments of the constructor. Additionally, some of the
      items have getters that are used by the ``Workers`` that execute the
      experiments and by the training loop.
    * the training algorithm, with the main entry point being ``train()``.

    Note:
        When executing a ``BasicTraining`` in the cloud, additional constraints
        are applied to the data. For example, the model is restricted to
        sequential layers of specific types; the dataset choices are limited,
        etc. Please check the ``CloudRunner`` documentation.
    """

    def __init__(
            self,
            dataset: Type[Dataset],
            model: Module,
            batch_size: int = 64,
            loss_function: type = NLLLoss,
            epochs: int = 30,
            learning_rate: float = 0.05
    ):
        """Create a new ``BasicTraining``.

        Args:
            dataset: the dataset class to be used.
            model: the neural network to be trained.
            batch_size: the batch size used for training.
            loss_function: the loss function used for training.
            epochs: the number of epochs for the training.
            learning_rate: the learning rate used by the optimizer.
        """
        self.dataset = dataset
        self.model = model
        self.batch_size = batch_size
        self.loss_function = loss_function
        self.epochs = epochs
        self.learning_rate = learning_rate

        super().__init__()

    def get_dataset_arguments(self, dataset: type) -> Tuple[Dict, Dict]:
        """Return the dataset constructor arguments for specifying subset."""
        if dataset in (SVHN, ):
            return {'split': 'train'}, {'split': 'test'}
        return {'train': True}, {'train': False}

    def get_dataset_transform(self, dataset: type) -> Any:
        """Return the dataset transform."""
        # Normalize supported datasets.
        if dataset == FashionMNIST:
            mean = Tensor([0.2860])
            std = Tensor([0.3205])
            transform = Compose([ToTensor(), Normalize(mean, std)])
        elif dataset == SVHN:
            mean = Tensor([0.4377, 0.4438, 0.4728])
            std = Tensor([0.1980, 0.2010, 0.1970])
            transform = Compose([ToTensor(), Normalize(mean, std)])
        else:
            transform = Compose([ToTensor()])

        return transform

    def get_data_loaders(
            self,
            dataset: type,
            batch_size: int,
            max_elements_train: int = 0,
            dataset_root: str = '/tmp/datasets'
    ) -> Tuple[DataLoader, DataLoader]:
        """Return `DataLoaders` for the selected dataset.

        Args:
            dataset: the dataset class to be used.
            batch_size: the batch size used for training.
            max_elements_train: the maximum number of elements of the dataset
                to be used. If ``0``, the full dataset is used.
            dataset_root: the path to the folder where the files from the
                dataset are stored.

        Returns:
            A tuple with the training and validation loaders.
        """
        # Create the sets and the loaders.
        train_args, test_args = self.get_dataset_arguments(dataset)
        transform = self.get_dataset_transform(dataset)

        # Create the datasets.
        training_set = dataset(dataset_root, transform=transform, **train_args)
        validation_set = dataset(dataset_root, transform=transform, **test_args)

        if max_elements_train:
            training_set = Subset(training_set, range(max_elements_train))
            validation_set = Subset(validation_set, range(max_elements_train))

        training_loader = DataLoader(training_set, batch_size=batch_size, shuffle=True)
        validation_loader = DataLoader(validation_set, batch_size=batch_size, shuffle=True)

        return training_loader, validation_loader

    def get_optimizer(
            self,
            learning_rate: float,
            model: Module
    ) -> Optimizer:
        """Return the `Optimizer` for the experiment.

        Args:
            learning_rate: the learning rate used by the optimizer.
            model: the neural network to be trained.

        Returns:
            the optimizer to be used in the experiment.
        """
        optimizer = AnalogSGD(model.parameters(), lr=learning_rate)
        optimizer.regroup_param_groups(model)

        return optimizer

    def training_step(
            self,
            training_loader: DataLoader,
            model: Module,
            optimizer: Optimizer,
            loss_function: _Loss,
            device: torch_device
    ) -> None:
        """Run a single training step.

        Args:
            training_loader: the data loader for the training data.
            model: the neural network to be trained.
            optimizer: the optimizer used for the training.
            loss_function: the loss function used for training.
            device: the torch device used for the model.
        """
        model.train()

        for images, labels in training_loader:
            # Move the data to the device if needed.
            if device:
                images = images.to(device)
                labels = labels.to(device)

            optimizer.zero_grad()
            output = model(images)
            loss = loss_function(output, labels)
            batch_image_count = labels.size(0)

            # Run training (backward propagation).
            loss.backward()

            # Optimize weights.
            optimizer.step()

            self._call_hook(Signals.TRAIN_EPOCH_BATCH_END,
                            batch_image_count,
                            loss.item()*batch_image_count)

    def validation_step(
            self,
            validation_loader: DataLoader,
            model: Module,
            loss_function: _Loss,
            device: torch_device
    ) -> None:
        """Run a single evaluation step.

        Args:
            validation_loader: the data loader for the validation data.
            model: the neural network to be trained.
            loss_function: the loss function used for training.
            device: the torch device used for the model.
        """
        model.eval()

        for images, labels in validation_loader:
            # Move the data to the device if needed.
            if device:
                images = images.to(device)
                labels = labels.to(device)

            # Predict image.
            prediction = model(images)
            loss = loss_function(prediction, labels)

            _, predicted = torch_max(prediction.data, 1)
            batch_image_count = labels.size(0)
            batch_correct_count = (predicted == labels).sum().item()

            self._call_hook(Signals.VALIDATION_EPOCH_BATCH_END,
                            batch_image_count,
                            batch_correct_count,
                            loss.item()*batch_image_count)

    def train(
            self,
            training_loader: DataLoader,
            validation_loader: DataLoader,
            model: Module,
            optimizer: Optimizer,
            loss_function: _Loss,
            epochs: int,
            device: torch_device
    ) -> List[Dict]:
        """Run the training loop.

        Args:
            training_loader: the data loader for the training data.
            validation_loader: the data loader for the validation data.
            model: the neural network to be trained.
            optimizer: the optimizer used for the training.
            loss_function: the loss function used for training.
            epochs: the number of epochs for the training.
            device: the torch device used for the model.

        Returns:
            A list of the metrics for each epoch.
        """
        results = []

        # Move the model to the device if needed.
        if device:
            model = model.to(device)

        for epoch_number in range(epochs):
            self._call_hook(Signals.EPOCH_START, epoch_number)
            self._call_hook(Signals.TRAIN_EPOCH_START, epoch_number)
            self.training_step(training_loader,
                               model,
                               optimizer,
                               loss_function,
                               device)
            self._call_hook(Signals.TRAIN_EPOCH_END)

            self._call_hook(Signals.VALIDATION_EPOCH_START, epoch_number)
            self.validation_step(validation_loader,
                                 model,
                                 loss_function,
                                 device)
            self._call_hook(Signals.VALIDATION_EPOCH_END)

            epoch_results = {'epoch': epoch_number}
            epoch_results.update(self._call_hook(Signals.EPOCH_END))
            results.append(epoch_results)

        return results

    def __str__(self) -> str:
        """Return a string representation of a BasicTraining experiment."""
        return ('{}(dataset={}, batch_size={}, loss_function={}, epochs={}, '
                'learning_rate={}, model={})'.format(
                    self.__class__.__name__,
                    getattr(self.dataset, '__name__', self.dataset),
                    self.batch_size,
                    getattr(self.loss_function, '__name__', self.loss_function),
                    self.epochs,
                    self.learning_rate,
                    self.model
                ))



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Helper for retrieving Metrics of an Experiment."""

from datetime import datetime
from typing import Dict, List


class LocalMetric:
    """Metric for local experiments.

    Metric for local execution of experiments. Output to stdout can be
    controlled by the ``stdout`` parameter to the constructor.
    """

    def __init__(self, stdout: bool = True) -> None:
        self.epochs: List[Dict] = []
        self.current_epoch: Dict = {}
        self.stdout = stdout

    def receive_epoch_start(self, epoch: int) -> None:
        """Hook for `EPOCH_START`."""
        self.current_epoch = {
            'number': epoch,
            'start_time': datetime.utcnow(),
            'total_loss': 0,
            'training_images': 0,
            'validation_images': 0,
            'validation_correct': 0,
            'validation_loss': 0
        }

    def receive_train_epoch_batch_end(
            self,
            total: int,
            train_loss: float
    ) -> None:
        """Hook for `TRAIN_EPOCH_START`."""
        self.current_epoch['training_images'] += total
        self.current_epoch['total_loss'] += train_loss

    def receive_validation_epoch_batch_end(
            self,
            total: int,
            correct: int,
            validation_loss: float
    ) -> None:
        """Hook for `VALIDATION_EPOCH_BATCH_END`."""
        self.current_epoch['validation_images'] += total
        self.current_epoch['validation_correct'] += int(correct)
        self.current_epoch['validation_loss'] += validation_loss

    def receive_train_epoch_end(self) -> None:
        """Hook for `TRAIN_EPOCH_END`."""
        if not self.stdout:
            return

        print('Epoch: {}, loss: {:.8f}'.format(
                self.current_epoch['number'],
                self.current_epoch['total_loss'] / self.current_epoch['training_images'],
              ))

    def receive_validation_epoch_end(self) -> None:
        """Hook for `VALIDATION_EPOCH_END`."""
        if not self.stdout:
            return

        print('Number of images: {}, accuracy: {:.6%}, validation loss: {:.8f}'.format(
            self.current_epoch['validation_images'],
            self.current_epoch['validation_correct'] / self.current_epoch['validation_images'],
            self.current_epoch['validation_loss'] / self.current_epoch['validation_images'],
        ))

    def receive_epoch_end(self) -> Dict:
        """Hook for `EPOCH_END`."""
        end_time = datetime.utcnow()
        time_epoch = (end_time - self.current_epoch['start_time']).total_seconds()
        if self.stdout:
            print('Time for epoch {}: {:.4}s'.format(self.current_epoch['number'],
                                                     time_epoch))
        self.epochs.append(self.current_epoch)

        return {
            'epoch': self.current_epoch['number'],
            'time_epoch': time_epoch,
            'accuracy': (self.current_epoch['validation_correct'] /
                         self.current_epoch['validation_images']),
            'train_loss': (self.current_epoch['total_loss'] /
                           self.current_epoch['training_images']),
            'valid_loss': (self.current_epoch['validation_loss'] /
                           self.current_epoch['validation_images'])
        }



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test for different utility functionality."""

from tempfile import TemporaryFile

from numpy import array
from numpy.random import rand
from numpy.testing import assert_array_almost_equal, assert_raises
from torch import Tensor, load, save
from torch.nn import Module, Sequential
from torch.nn.functional import mse_loss

from aihwkit.nn import AnalogConv2d
from aihwkit.optim import AnalogSGD
from aihwkit.simulator.configs import SingleRPUConfig
from aihwkit.simulator.configs.devices import ConstantStepDevice, LinearStepDevice
from aihwkit.simulator.configs.utils import IOParameters, UpdateParameters
from aihwkit.exceptions import TileError

from .helpers.decorators import parametrize_over_layers
from .helpers.layers import Conv2d, Conv2dCuda, Linear, LinearCuda
from .helpers.testcases import ParametrizedTestCase
from .helpers.tiles import FloatingPoint, ConstantStep, Inference


@parametrize_over_layers(
    layers=[Linear, Conv2d, LinearCuda, Conv2dCuda],
    tiles=[FloatingPoint, ConstantStep, Inference],
    biases=[True, False]
)
class SerializationTest(ParametrizedTestCase):
    """Tests for serialization."""

    @staticmethod
    def train_model(model, loss_func, x_b, y_b):
        """Train the model."""
        opt = AnalogSGD(model.parameters(), lr=0.5)
        opt.regroup_param_groups(model)

        epochs = 1
        for _ in range(epochs):
            pred = model(x_b)
            loss = loss_func(pred, y_b)

            loss.backward()
            opt.step()
            opt.zero_grad()

    @staticmethod
    def get_layer_and_tile_weights(model):
        """Return the weights and biases of the model and the tile."""
        weight = model.weight.data.detach().cpu().numpy()
        if model.use_bias:
            bias = model.bias.data.detach().cpu().numpy()
        else:
            bias = None

        analog_weight, analog_bias = model.analog_tile.get_weights()
        analog_weight = analog_weight.detach().cpu().numpy().reshape(weight.shape)
        if model.use_bias:
            analog_bias = analog_bias.detach().cpu().numpy()
        else:
            analog_bias = None

        return weight, bias, analog_weight, analog_bias

    def test_save_load_state_dict_train(self):
        """Test saving and loading using a state dict after training."""
        model = self.get_layer()

        # Perform an update in order to modify tile weights and biases.
        loss_func = mse_loss
        if isinstance(model, AnalogConv2d):
            input_x = Tensor(rand(2, 2, 3, 3))*0.2
            input_y = Tensor(rand(2, 3, 4, 4))*0.2
        else:
            input_x = Tensor(rand(2, model.in_features))*0.2
            input_y = Tensor(rand(2, model.out_features))*0.2

        if self.use_cuda:
            input_x = input_x.cuda()
            input_y = input_y.cuda()

        self.train_model(model, loss_func, input_x, input_y)

        # Keep track of the current weights and biases for comparing.
        (model_weights, model_biases,
         tile_weights, tile_biases) = self.get_layer_and_tile_weights(model)

        # now the tile weights should be out of sync
        assert_raises(AssertionError, assert_array_almost_equal, model_weights, tile_weights)
        if self.bias:
            assert_raises(AssertionError, assert_array_almost_equal, model_biases, tile_biases)

        # Save the model to a file.
        with TemporaryFile() as file:
            save(model.state_dict(), file)
            # Create a new model and load its state dict.
            file.seek(0)
            new_model = self.get_layer()
            new_model.load_state_dict(load(file))

        # Compare the new model weights and biases. they should now be in sync
        (new_model_weights, new_model_biases,
         new_tile_weights, new_tile_biases) = self.get_layer_and_tile_weights(new_model)

        assert_array_almost_equal(tile_weights, new_model_weights)
        assert_array_almost_equal(tile_weights, new_tile_weights)
        if self.bias:
            assert_array_almost_equal(tile_biases, new_model_biases)
            assert_array_almost_equal(tile_biases, new_tile_biases)

    def test_save_load_model(self):
        """Test saving and loading a model directly."""
        model = self.get_layer()

        # Keep track of the current weights and biases for comparing.
        (model_weights, model_biases,
         tile_weights, tile_biases) = self.get_layer_and_tile_weights(model)
        assert_array_almost_equal(model_weights, tile_weights)
        if self.bias:
            assert_array_almost_equal(model_biases, tile_biases)

        # Save the model to a file.
        with TemporaryFile() as file:
            save(model, file)
            # Load the model.
            file.seek(0)
            new_model = load(file)

        # Compare the new model weights and biases.
        (new_model_weights, new_model_biases,
         new_tile_weights, new_tile_biases) = self.get_layer_and_tile_weights(new_model)

        assert_array_almost_equal(model_weights, new_model_weights)
        assert_array_almost_equal(tile_weights, new_tile_weights)
        if self.bias:
            assert_array_almost_equal(model_biases, new_model_biases)
            assert_array_almost_equal(tile_biases, new_tile_biases)

        # Asserts over the AnalogContext of the new model.
        self.assertTrue(hasattr(new_model.analog_tile.analog_ctx, 'analog_tile'))
        self.assertIsInstance(new_model.analog_tile.analog_ctx.analog_tile,
                              model.analog_tile.__class__)

    def test_save_load_meta_parameter(self):
        """Test saving and loading a device with custom parameters."""
        # Create the device and the array.
        rpu_config = SingleRPUConfig(
            forward=IOParameters(inp_noise=0.321),
            backward=IOParameters(inp_noise=0.456),
            update=UpdateParameters(desired_bl=78),
            device=ConstantStepDevice(w_max=0.987)
        )

        model = self.get_layer(rpu_config=rpu_config)

        # Save the model to a file.
        with TemporaryFile() as file:
            save(model, file)
            # Load the model.
            file.seek(0)
            new_model = load(file)

        # Assert over the new model tile parameters.
        parameters = new_model.analog_tile.tile.get_parameters()
        self.assertAlmostEqual(parameters.forward_io.inp_noise, 0.321)
        self.assertAlmostEqual(parameters.backward_io.inp_noise, 0.456)
        self.assertAlmostEqual(parameters.update.desired_bl, 78)

    def test_save_load_hidden_parameters(self):
        """Test saving and loading a device with hidden parameters."""
        # Create the device and the array.
        model = self.get_layer()
        hidden_parameters = model.analog_tile.tile.get_hidden_parameters()

        # Save the model to a file.
        with TemporaryFile() as file:
            save(model, file)
            # Load the model.
            file.seek(0)
            new_model = load(file)

        # Assert over the new model tile parameters.
        new_hidden_parameters = new_model.analog_tile.tile.get_hidden_parameters()
        assert_array_almost_equal(hidden_parameters, new_hidden_parameters)

    def test_save_load_alpha_scale(self):
        """Test saving and loading a device with alpha_scale."""
        # Create the device and the array.
        model = self.get_layer()
        alpha = 2.0
        model.analog_tile.tile.set_alpha_scale(alpha)

        # Save the model to a file.
        with TemporaryFile() as file:
            save(model, file)
            # Load the model.
            file.seek(0)
            new_model = load(file)

        # Assert over the new model tile parameters.
        alpha_new = new_model.analog_tile.tile.get_alpha_scale()
        assert_array_almost_equal(array(alpha), array(alpha_new))

    def test_save_load_weight_scaling_omega(self):
        """Test saving and loading a device with weight scaling omega."""
        model = self.get_layer(weight_scaling_omega=0.5)

        alpha = model.analog_tile.tile.get_alpha_scale()
        self.assertNotEqual(alpha, 1.0)

        # Save the model to a file.
        with TemporaryFile() as file:
            save(model, file)
            # Load the model.
            file.seek(0)
            new_model = load(file)

        # Assert over the new model tile parameters.
        alpha_new = new_model.analog_tile.tile.get_alpha_scale()
        assert_array_almost_equal(array(alpha), array(alpha_new))

    def test_save_load_state_dict_hidden_parameters(self):
        """Test saving and loading via state_dict with hidden parameters."""
        # Create the device and the array.
        model = self.get_layer()
        hidden_parameters = model.analog_tile.tile.get_hidden_parameters()

        # Save the model to a file.
        with TemporaryFile() as file:
            save(model.state_dict(), file)
            # Load the model.
            file.seek(0)
            new_model = self.get_layer()
            new_model.load_state_dict(load(file))

        # Assert over the new model tile parameters.
        new_hidden_parameters = new_model.analog_tile.tile.get_hidden_parameters()
        assert_array_almost_equal(hidden_parameters, new_hidden_parameters)

    def test_state_dict_children_layers_sequential(self):
        """Test using the state_dict with children analog layers via Sequential."""
        children_layer = self.get_layer()
        model = Sequential(children_layer)

        # Keep track of the current weights and biases for comparing.
        (model_weights, model_biases,
         tile_weights, tile_biases) = self.get_layer_and_tile_weights(children_layer)

        self.assertIn('0.analog_tile_state', model.state_dict())

        # Update the state_dict of a new model.
        new_children_layer = self.get_layer()
        new_model = Sequential(new_children_layer)
        new_model.load_state_dict(model.state_dict())

        # Compare the new model weights and biases.
        (new_model_weights, new_model_biases, new_tile_weights, new_tile_biases) = \
            self.get_layer_and_tile_weights(new_children_layer)

        assert_array_almost_equal(model_weights, new_model_weights)
        assert_array_almost_equal(tile_weights, new_tile_weights)
        if self.bias:
            assert_array_almost_equal(model_biases, new_model_biases)
            assert_array_almost_equal(tile_biases, new_tile_biases)

    def test_state_dict_children_layers_subclassing(self):
        """Test using the state_dict with children analog layers via subclassing."""

        class CustomModule(Module):
            """Module that defines its children layers via custom attributes."""
            # pylint: disable=abstract-method
            def __init__(self, layer: Module):
                super().__init__()
                self.custom_child = layer

        children_layer = self.get_layer()
        model = CustomModule(children_layer)

        # Keep track of the current weights and biases for comparing.
        (model_weights, model_biases, tile_weights, tile_biases) = \
            self.get_layer_and_tile_weights(children_layer)

        self.assertIn('custom_child.analog_tile_state', model.state_dict())

        # Update the state_dict of a new model.
        new_children_layer = self.get_layer()
        new_model = CustomModule(new_children_layer)
        new_model.load_state_dict(model.state_dict())

        # Compare the new model weights and biases.
        (new_model_weights, new_model_biases, new_tile_weights, new_tile_biases) = \
            self.get_layer_and_tile_weights(new_children_layer)

        assert_array_almost_equal(model_weights, new_model_weights)
        assert_array_almost_equal(tile_weights, new_tile_weights)
        if self.bias:
            assert_array_almost_equal(model_biases, new_model_biases)
            assert_array_almost_equal(tile_biases, new_tile_biases)

    def test_state_dict_analog_strict(self):
        """Test the `strict` flag for analog layers."""
        model = self.get_layer()
        state_dict = model.state_dict()

        # Remove the analog key from the state dict.
        del state_dict['analog_tile_state']

        # Check that it fails when using `strict`.
        with self.assertRaises(RuntimeError) as context:
            model.load_state_dict(state_dict, strict=True)
        self.assertIn('Missing key', str(context.exception))

        # Check that it passes when not using `strict`.
        model.load_state_dict(state_dict, strict=False)

    def test_state_dict(self):
        """Test creating a new model using a state dict, without saving to disk."""
        model = self.get_layer()
        state_dict = model.state_dict()

        new_model = self.get_layer()
        new_model.load_state_dict(state_dict)

        # Asserts over the AnalogContext of the new model.
        self.assertTrue(hasattr(new_model.analog_tile.analog_ctx, 'analog_tile'))
        self.assertIsInstance(new_model.analog_tile.analog_ctx.analog_tile,
                              model.analog_tile.__class__)

    def test_hidden_parameter_mismatch(self):
        """Test for error if tile structure mismatches."""
        model = self.get_layer()
        state_dict = model.state_dict()

        # Create the device and the array.
        rpu_config = SingleRPUConfig(
            device=LinearStepDevice()  # different hidden structure
        )

        new_model = self.get_layer(rpu_config=rpu_config)
        if new_model.analog_tile.__class__.__name__ != model.analog_tile.__class__.__name__:
            with self.assertRaises(TileError):
                self.assertRaises(new_model.load_state_dict(state_dict))



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""High level analog tiles (base)."""

from collections import OrderedDict
from typing import Dict, Generic, List, Optional, Tuple, TypeVar, Union

from numpy import concatenate, expand_dims
from numpy import abs as numpy_abs
from torch import Tensor, stack
from torch import device as torch_device
from torch.autograd import no_grad

from aihwkit.exceptions import TileError
from aihwkit.simulator.rpu_base import tiles
from aihwkit.optim.context import AnalogContext

RPUConfigGeneric = TypeVar('RPUConfigGeneric')


class BaseTile(Generic[RPUConfigGeneric]):
    """Base class for tiles.

    Args:
        out_size: output size
        in_size: input size
        rpu_config: resistive processing unit configuration.
        bias: whether to add a bias column to the tile.
        in_trans: Whether to assume an transposed input (batch first)
        out_trans: Whether to assume an transposed output (batch first)
    """
    # pylint: disable=too-many-instance-attributes,too-many-public-methods

    def __init__(
            self,
            out_size: int,
            in_size: int,
            rpu_config: RPUConfigGeneric,
            bias: bool = True,
            in_trans: bool = False,
            out_trans: bool = False
    ):
        self.out_size = out_size
        self.in_size = in_size
        self.rpu_config = rpu_config
        self.bias = bias
        self.in_trans = in_trans
        self.out_trans = out_trans

        # Only used for indexed.
        self.image_sizes = []  # type: List[int]

        x_size = in_size + 1 if self.bias else in_size
        d_size = out_size

        self.tile = self._create_simulator_tile(x_size, d_size, rpu_config)
        self.tile.set_learning_rate(0.01)
        self.tile.set_weights_uniform_random(-0.01, 0.01)

        self.device = torch_device('cpu')
        self.is_cuda = False
        self.shared_weights = None  # type: Optional[Tensor]

        # create analog context
        self.analog_ctx = AnalogContext(self)

    @no_grad()
    def get_analog_ctx(self) -> AnalogContext:
        """Return the analog context of the tile to be used in ``AnalogFunction``."""
        return self.analog_ctx

    @no_grad()
    def ensure_shared_weights(self, shared_weights: Optional[Tensor] = None) -> None:
        """Ensure that the shared_weights is set properly.

        No-op if shared weights is not used.
        """
        if shared_weights is not None:
            self.shared_weights.data = shared_weights.data  # type: ignore[union-attr]

        if self.shared_weights is not None:
            self.tile.set_shared_weights(self.shared_weights.data)

    @no_grad()
    def set_delta_weights(self, delta_weights: Optional[Tensor] = None) -> None:
        """Set the weight grad tensor and set the update to.

        No-op if shared weights is not used.
        """
        if self.shared_weights is not None and delta_weights is not None:
            self.tile.set_delta_weights(delta_weights)

    @no_grad()
    def reset_delta_weights(self) -> None:
        """Reset the weight grad tensor to default update behavior (i.e. adding the
        update directly to the weight).

        No-op if shared weights is not used.
        """
        if self.shared_weights is not None:
            self.tile.reset_delta_weights()

    @no_grad()
    def get_brief_info(self) -> str:
        """Return short info about the underlying C++ tile."""
        return self.tile.get_brief_info().rstrip()

    def __getstate__(self) -> Dict:
        """Get the state for pickling.

        This method removes the ``tile`` member, as the binding Tiles are not
        serializable.
        """
        current_dict = self.__dict__.copy()

        current_dict['analog_tile_weights'] = self.tile.get_weights()
        # Store the hidden parameters as a numpy array, as storing it as
        # Tensor causes issues in PyTorch 1.5.
        current_dict['analog_tile_hidden_parameters'] \
            = self.tile.get_hidden_parameters().data.numpy()
        current_dict['analog_tile_hidden_parameter_names'] \
            = self.tile.get_hidden_parameter_names()
        current_dict['analog_tile_class'] = self.__class__.__name__
        current_dict['analog_alpha_scale'] = self.tile.get_alpha_scale()
        current_dict.pop('tile')
        current_dict.pop('stream', None)

        return current_dict

    def __setstate__(self, state: Dict) -> None:
        """Set the state after unpickling.

        This method recreates the ``tile`` member, creating a new one from
        scratch, as the binding Tiles are not serializable.

        Caution:
            RPU configs are overwritten by loading the state.

        Raises:
            TileError: if tile class does not match or hidden parameters do not match
        """
        current_dict = state.copy()
        weights = current_dict.pop('analog_tile_weights')
        hidden_parameters = current_dict.pop('analog_tile_hidden_parameters')
        hidden_parameters_names = current_dict.pop('analog_tile_hidden_parameter_names', [])
        alpha_scale = current_dict.pop('analog_alpha_scale')
        tile_class = current_dict.pop('analog_tile_class', self.__class__.__name__)
        self.__dict__.update(current_dict)

        x_size = self.in_size + 1 if self.bias else self.in_size
        d_size = self.out_size

        # Recreate the tile.
        # Check for tile mismatch
        if tile_class != self.__class__.__name__:
            raise TileError(
                'Mismatch of tile class: {} versus {}. Can only load analog '
                'state from the same tile class.'.format(self.__class__.__name__, tile_class))

        self.tile = self._create_simulator_tile(x_size, d_size, self.rpu_config)
        names = self.tile.get_hidden_parameter_names()
        if len(hidden_parameters_names) > 0 and names != hidden_parameters_names:
            # Check whether names match
            raise TileError('Mismatch with loaded analog state: '
                            'Hidden parameter structure is unexpected.')
        self.tile.set_hidden_parameters(Tensor(hidden_parameters))
        self.tile.set_weights(weights)
        if alpha_scale is not None:
            self.tile.set_alpha_scale(alpha_scale)

        # Keep the data (for future use)
        data = self.analog_ctx.data.detach()
        self.analog_ctx = AnalogContext(self)
        self.analog_ctx.set_data(data)

        if self.is_cuda:
            self.cuda(self.device)

        self.ensure_shared_weights()

    def _create_simulator_tile(
            self,
            x_size: int,
            d_size: int,
            rpu_config: RPUConfigGeneric
    ) -> Union[tiles.FloatingPointTile, tiles.AnalogTile]:
        """Create a simulator tile.

        Args:
            x_size: input size
            d_size: output size
            rpu_config: resistive processing unit configuration

        Returns:
            a simulator tile based on the specified configuration.
        """
        raise NotImplementedError

    def set_weights(
            self,
            weights: Tensor,
            biases: Optional[Tensor] = None,
            realistic: bool = False,
            n_loops: int = 10
    ) -> None:
        """Set the tile weights (and biases).

        Sets the internal tile weights to the specified values, and also the
        internal tile biases if the tile was set to use bias (via
        ``self.bias``).

        Note:
            By default this is **not** hardware realistic. You can set the
            ``realistic`` parameter to ``True`` for a realistic transfer.

        Args:
            weights: ``[out_size, in_size]`` weight matrix.
            biases: ``[out_size]`` bias vector. This parameter is required if
                ``self.bias`` is ``True``, and ignored otherwise.
            realistic: whether to use the forward and update pass to
                program the weights iteratively, using
                :meth:`set_weights_realistic`.
            n_loops: number of times the columns of the weights are set in a
                closed-loop manner.
                A value of ``1`` means that all columns in principle receive
                enough pulses to change from ``w_min`` to ``w_max``.

        Returns:
            None.

        Raises:
            ValueError: if the tile has bias but ``bias`` has not been
                specified.
        """
        # Prepare the array expected by the pybind function, appending the
        # biases row if needed.
        weights_numpy = weights.clone().detach().cpu().numpy()

        if self.bias:
            # Create a ``[out_size, in_size (+ 1)]`` matrix.
            if biases is None:
                raise ValueError('Analog tile has a bias, but no bias given')

            biases_numpy = expand_dims(biases.clone().detach().cpu().numpy(), 1)
            combined_weights = concatenate([weights_numpy, biases_numpy], axis=1)
        else:
            # Use only the ``[out_size, in_size]`` matrix.
            combined_weights = weights_numpy

        if realistic:
            return self.tile.set_weights_realistic(combined_weights, n_loops)

        return self.tile.set_weights(combined_weights)

    def get_weights(
            self,
            realistic: bool = False
    ) -> Tuple[Tensor, Optional[Tensor]]:
        """Get the tile weights (and biases).

        Gets the tile weights and extracts the mathematical weight
        matrix and biases (if present, by determined by the ``self.bias``
        parameter).

        Note:
            By default this is **not** hardware realistic. Use set
            ``realistic`` to True for a realistic transfer.

        Args:
            realistic: Whether to use the forward pass to read out the tile
                weights iteratively, using :meth:`get_weights_realistic`.

        Returns:
            a tuple where the first item is the ``[out_size, in_size]`` weight
            matrix; and the second item is either the ``[out_size]`` bias vector
            or ``None`` if the tile is set not to use bias.
        """
        # Retrieve the internal weights (and potentially biases) matrix.
        if realistic:
            combined_weights = self.tile.get_weights_realistic()
        else:
            combined_weights = self.tile.get_weights()

        # Split the internal weights (and potentially biases) matrix.
        if self.bias:
            # combined_weights is [out_size, in_size (+ 1)].
            weights = Tensor(combined_weights[:, :-1])
            biases = Tensor(combined_weights[:, -1])
        else:
            # combined_weights is [out_size, in_size].
            weights = Tensor(combined_weights)
            biases = None

        return weights, biases if self.bias else None

    def set_weights_scaled(
            self,
            weights: Tensor,
            biases: Optional[Tensor] = None,
            realistic: bool = False,
            n_loops: int = 10,
            omega: float = 1.0
    ) -> None:
        r"""Set the tile weights (and biases) in a scaled fashion.

        Similar to :meth:`set_weights`, however, additionally scales the weights
        by a global scale :math:`\alpha`, that is then applied in digital at the
        output of forward and backward pass, and the learning rate for this tile
        is adjusted accordingly.

        The weights are scaled by :math:`\omega/\max_{ij} |w_{ij}|` and the global
        digital factor :math:`alpha` is set to :math:`\max_{ij} |w_{ij}|/\omega`.

        It can be shown that such a constant factor greatly improves the SNR and
        training accuracy as the full weight range of the analog devices are
        used. See also `Rasch, Gokmen & Haensch (2019)`_ for more details.

        Caution:
            Using ``get_weights`` will now retrieve the true analog weights
            *without* applying the global factor. To get the true weights, use
            ``get_weights`` and scale it by the :math:`\alpha` of this layer
            which can be retrieved by ``get_alpha_scale()``.

        Args:
            weights: ``[out_size, in_size]`` weight matrix.
            biases: ``[out_size]`` bias vector. This parameter is required if
                ``self.bias`` is ``True``, and ignored otherwise.
            realistic: whether to use the forward and update pass to program the
                weights iteratively, using :meth:`set_weights_realistic`.
            n_loops: number of times the columns of the weights are set in a
                closed-loop manner.
                A value of ``1`` means that all columns in principle receive
                enough pulses to change from ``w_min`` to ``w_max``.
            omega: where the weight max should be mapped in terms of the weight
                range. Note that for ``omega`` larger than the maximal weight of
                the device, weights will get clipped for most devices.

        Returns:
            None.

        Raises:
            ValueError: if the tile has bias but ``bias`` has not been
                specified.

        .. _`Rasch, Gokmen & Haensch (2019)`: https://arxiv.org/abs/1906.02698
        """
        # Prepare the array expected by the pybind function, appending the
        # biases row if needed.
        weights_numpy = weights.clone().detach().cpu().numpy()

        if self.bias:
            # Create a ``[out_size, in_size (+ 1)]`` matrix.
            if biases is None:
                raise ValueError('Analog tile has a bias, but no bias given')

            biases_numpy = expand_dims(biases.clone().detach().cpu().numpy(), 1)
            combined_weights = concatenate([weights_numpy, biases_numpy], axis=1)
        else:
            # Use only the ``[out_size, in_size]`` matrix.
            combined_weights = weights_numpy

        # Scale the weights.
        weight_max = numpy_abs(combined_weights).max()
        combined_weights = combined_weights/weight_max*omega
        alpha = weight_max/omega
        self.tile.set_alpha_scale(alpha)

        if realistic:
            return self.tile.set_weights_realistic(combined_weights, n_loops)

        return self.tile.set_weights(combined_weights)

    def get_weights_scaled(
            self,
            realistic: bool = False
    ) -> Tuple[Tensor, Optional[Tensor]]:
        """Get the tile weights (and biases) and applies the current alpha
        scale to it.

        Gets the tile weights and extracts the mathematical weight
        matrix and biases (if present, by determined by the ``self.bias``
        parameter).

        Note:
            By default this is **not** hardware realistic. Use set
            ``realistic`` to True for a realistic transfer.

        Args:
            realistic: Whether to use the forward pass to read out the tile
                weights iteratively, using :meth:`get_weights_realistic`.

        Returns:
            a tuple where the first item is the ``[out_size, in_size]`` weight
            matrix; and the second item is either the ``[out_size]`` bias vector
            or ``None`` if the tile is set not to use bias. Both have the alpha
            scale applied.
        """
        weights, biases = self.get_weights(realistic=realistic)
        alpha = self.tile.get_alpha_scale()

        if self.bias:
            return weights*alpha, biases*alpha

        return weights*alpha, None

    def set_learning_rate(self, learning_rate: float) -> None:
        """Set the tile learning rate.

        Set the tile learning rate to ``-learning_rate``. Note that the
        learning rate is always taken to be negative (because of the meaning in
        gradient descent) and positive learning rates are not supported.

        Args:
            learning_rate: the desired learning rate.

        Returns:
            None.
        """
        return self.tile.set_learning_rate(learning_rate)

    def get_learning_rate(self) -> float:
        """Return the tile learning rate.

        Returns:
            float: the tile learning rate.
        """
        return self.tile.get_learning_rate()

    @no_grad()
    def decay_weights(self, alpha: float = 1.0) -> None:
        """Decays the weights once according to the decay parameters of the tile.

        Args:
            alpha: additional decay scale (such as LR). The base decay
                rate is set during tile init.

        Returns:
            None.
        """
        return self.tile.decay_weights(alpha)

    @no_grad()
    def drift_weights(self, delta_t: float = 1.0) -> None:
        """Drifts the weights once according to the drift parameters of the
        tile.

        See also :class:`~aihwkit.simulator.configs.utils.DriftParameter`.

        Args:
            delta_t: Time since last drift call.

        Returns:
            None.
        """
        return self.tile.drift_weights(delta_t)

    @no_grad()
    def diffuse_weights(self) -> None:
        """Diffuses the weights once according to the diffusion parameters of
        the tile.

        The base diffusion rate is set during tile init.

        Returns:
            None
        """
        return self.tile.diffuse_weights()

    @no_grad()
    def reset_columns(
            self,
            start_column_idx: int = 0,
            num_columns: int = 1,
            reset_prob: float = 1.0
    ) -> None:
        r"""Reset (a number of) columns according to the reset parameters of the tile.

        Resets the weights with device-to-device and cycle-to-cycle
        variability (depending on device type), typically:

        .. math::
            W_{ij} = \xi*\sigma_\text{reset} + b^\text{reset}_{ij}

        The reset parameters are set during tile init.

        Args:
            start_column_idx: a start index of columns (0..x_size-1)
            num_columns: how many consecutive columns to reset (with circular warping)
            reset_prob: individual probability of reset.

        Returns:
            None
        """
        return self.tile.reset_columns(start_column_idx, num_columns, reset_prob)

    def cpu(self) -> 'BaseTile':
        """Return a copy of this tile in CPU memory."""
        raise NotImplementedError

    def cuda(
            self,
            device: Optional[Union[torch_device, str, int]] = None
    ) -> 'BaseTile':
        """Return a copy of this tile in CUDA memory."""
        raise NotImplementedError

    @no_grad()
    def forward(self, x_input: Tensor, is_test: bool = False) -> Tensor:
        """Perform the forward pass.

        Args:
            x_input: ``[N, in_size]`` tensor. If ``in_trans`` is set, transposed.
            is_test: whether to assume testing mode.

        Returns:
            torch.Tensor: ``[N, out_size]`` tensor. If ``out_trans`` is set, transposed.
        """
        # We use no-grad as we do it explicitly in the optimizer.
        return self.tile.forward(x_input, self.bias,
                                 self.in_trans, self.out_trans, is_test)

    def backward(self, d_input: Tensor) -> Tensor:
        """Perform the backward pass.

        Args:
            d_input: ``[N, out_size]`` tensor. If ``out_trans`` is set, transposed.

        Returns:
            torch.Tensor: ``[N, in_size]`` tensor. If ``in_trans`` is set, transposed.
        """
        return self.tile.backward(d_input, self.bias, self.out_trans, self.in_trans)

    def update(self, x_input: Tensor, d_input: Tensor) -> None:
        """Perform the update pass.

        Args:
            x_input: ``[N, in_size]`` tensor. If ``in_trans`` is set, transposed.
            d_input: ``[N, out_size]`` tensor. If ``out_trans`` is set, transposed.

        Returns:
            None
        """
        return self.tile.update(x_input, d_input, self.bias,
                                self.in_trans, self.out_trans)

    def get_hidden_parameters(self) -> OrderedDict:
        """Get the hidden parameters of the tile.

        Returns:
            Ordered dictionary of hidden parameter tensors.
        """
        names = self.tile.get_hidden_parameter_names()
        hidden_parameters = self.tile.get_hidden_parameters().detach_()

        ordered_parameters = OrderedDict()
        for idx, name in enumerate(names):
            ordered_parameters[name] = hidden_parameters[idx].clone()

        return ordered_parameters

    def set_hidden_parameters(self, ordered_parameters: OrderedDict) -> None:
        """Set the hidden parameters of the tile.

        Caution:
            Usually the hidden parameters are drawn according to the
            parameter definitions (those given in the RPU config). If
            the hidden parameters are arbitrary set by the user, then
            this correspondence might be broken. This might cause problems
            in the learning, in particular, the `weight granularity`
            (usually ``dw_min``, depending on the device) is needed for
            the dynamic adjustment of the bit length
            (``update_bl_management``, see
            :class:`~aihwkit.simulator.configs.utils.UpdateParameters`).

            Currently, the new ``dw_min`` parameter is tried to be
            estimated from the average of hidden parameters if the
            discrepancy with the ``dw_min`` from the definition is too
            large.

        Args:
            ordered_parameters: Ordered dictionary of hidden parameter tensors.

        Raises:
            TileError: In case the ordered dict keys do not conform
                with the current rpu config tile structure of the hidden
                parameters
        """
        if len(ordered_parameters) == 0:
            return

        hidden_parameters = stack(list(ordered_parameters.values()), dim=0)
        names = self.tile.get_hidden_parameter_names()
        if names != list(ordered_parameters.keys()):
            raise TileError('Mismatch with loaded analog state:'
                            'Hidden parameter structure is unexpected.')

        self.tile.set_hidden_parameters(hidden_parameters)

    def get_hidden_update_index(self) -> int:
        """Get the current updated device index of the hidden devices.

        Usually this is 0 as only one device is present per
        cross-point for many tile RPU configs. However, some RPU
        configs maintain internally multiple devices per cross-point
        (e.g. :class:`~aihwkit.simulator.config.devices.VectorUnitCell`).

        Returns:
            The next mini-batch updated device index.

        Note:
            Depending on the update and learning policy implemented
            in the tile, updated devices might switch internally as
            well.
        """
        return self.tile.get_hidden_update_index()

    def set_hidden_update_index(self, index: int) -> None:
        """Set the current updated hidden device index.

        Usually this is ignored and fixed to 0 as only one device is
        present per cross-point. Other devices, might not allow
        explicit setting as it would interfere with the implemented
        learning However rule. However, some tiles have internally
        multiple devices per cross-point (eg. unit cell) that can be
        chosen depending on the update policy.

        Args:
            index: device index to be updated in the next mini-batch

        Note:
            Depending on the update and learning policy implemented
            in the tile, updated devices might switch internally as
            well.
        """
        self.tile.set_hidden_update_index(index)

    def set_indexed(self, indices: Tensor, image_sizes: List) -> None:
        """Set the index matrix for convolutions ans switches to
        indexed forward/backward/update versions.

        Args:
            indices : torch.tensor with int indices
            image_sizes: [C_in, H_in, W_in, H_out, W_out] sizes

        Raises:
            ValueError: if ``image_sizes`` does not have valid dimensions.
            TileError: if the tile uses transposition.
        """
        if len(image_sizes) not in (3, 5, 7):
            raise ValueError('image_sizes expects 3, 5 or 7 sizes '
                             '[C_in, (D_in), H_in, (W_in), (D_out), H_out, (W_out)]')

        if self.in_trans or self.out_trans:
            raise TileError('Transposed indexed versions not supported (assumes NC(D)HW)')

        self.image_sizes = image_sizes
        self.tile.set_matrix_indices(indices)

    @no_grad()
    def forward_indexed(self, x_input: Tensor, is_test: bool = False) -> Tensor:
        """Perform the forward pass for convolutions.

        Depending on the input tensor size it performs the forward pass for a
        2D image or a 3D one.

        Args:
            x_input: ``[N, in_size]`` tensor. If ``in_trans`` is set, transposed.
            is_test: whether to assume testing mode.

        Returns:
            torch.Tensor: ``[N, out_size]`` tensor. If ``out_trans`` is set, transposed.

        Raises:
            TileError: if the indexed tile has not been initialized, or if
                ``self.images_sizes`` does not have a valid dimennion.
        """
        if not self.image_sizes:
            raise TileError('self.image_sizes is not initialized. Please use '
                            'set_indexed()')

        n_batch = x_input.size(0)
        channel_out = self.out_size

        if len(self.image_sizes) == 3:
            _, _, height_out = self.image_sizes
            d_tensor = x_input.new_empty((n_batch, channel_out, height_out))
        elif len(self.image_sizes) == 5:
            _, _, _, height_out, width_out = self.image_sizes
            d_tensor = x_input.new_empty((n_batch, channel_out, height_out, width_out))
        elif len(self.image_sizes) == 7:
            _, _, _, _, depth_out, height_out, width_out = self.image_sizes
            d_tensor = x_input.new_empty((n_batch, channel_out, depth_out, height_out, width_out))
        else:
            raise TileError('self.image_sizes length is not 3, 5 or 7')

        return self.tile.forward_indexed(x_input, d_tensor, is_test)

    def backward_indexed(self, d_input: Tensor) -> Tensor:
        """Perform the backward pass for convolutions.

        Depending on the input tensor size it performs the backward pass for a
        2D image or a 3D one.

        Args:
            d_input: ``[N, out_size]`` tensor. If ``out_trans`` is set, transposed.

        Returns:
            torch.Tensor: ``[N, in_size]`` tensor. If ``in_trans`` is set, transposed.

        Raises:
            TileError: if the indexed tile has not been initialized, or if
                ``self.images_sizes`` does not have a valid dimennion.
        """
        if not self.image_sizes:
            raise TileError('self.image_sizes is not initialized. Please use '
                            'set_indexed()')

        n_batch = d_input.size(0)

        if len(self.image_sizes) == 3:
            channel_in, height_in, _ = self.image_sizes
            x_tensor = d_input.new_empty((n_batch, channel_in, height_in))
        elif len(self.image_sizes) == 5:
            channel_in, height_in, width_in, _, _ = self.image_sizes
            x_tensor = d_input.new_empty((n_batch, channel_in, height_in, width_in))
        elif len(self.image_sizes) == 7:
            channel_in, depth_in, height_in, width_in, _, _, _ \
                = self.image_sizes
            x_tensor = d_input.new_empty((n_batch, channel_in, depth_in, height_in, width_in))
        else:
            raise TileError('self.image_sizes length is not 3, 5 or 7')

        return self.tile.backward_indexed(d_input, x_tensor)

    def update_indexed(self, x_input: Tensor, d_input: Tensor) -> None:
        """Perform the update pass for convolutions.

        Args:
            x_input: ``[N, in_size]`` tensor. If ``in_trans`` is set, transposed.
            d_input: ``[N, out_size]`` tensor. If ``out_trans`` is set, transposed.

        Returns:
            None
        """
        return self.tile.update_indexed(x_input, d_input)

    @no_grad()
    def post_update_step(self) -> None:
        """Operators that need to be called once per mini-batch."""
        if self.rpu_config.device.requires_diffusion():  # type: ignore
            self.diffuse_weights()
        if self.rpu_config.device.requires_decay():  # type: ignore
            self.decay_weights()



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Utilities for the AIHW Composer API."""

from os import getenv, path
from typing import Dict
from configparser import ConfigParser

DEFAULT_URL = 'https://api-aihw-composer.draco.res.ibm.com'


class ClientConfiguration:
    """Helper for retrieving the user configuration.

    Utility for retrieving the user configuration. The API token will be
    retrieved from, in order of preference:

    1. the ``AIHW_API_TOKEN`` environment variable.
    2. the ``aihwkit.conf`` configuration file.
    3. if the API token could not be found, it will return ``None``.

    The API URL will be retrieved from, in order of preference:

    1. the ``aihwkit.conf`` configuration file.
    2. the ``DEFAULT_URL`` module variable.

    The ``aihwkit.conf`` file is read from:

    * the current working directory.
    * ``XDG_CONFIG_HOME`` ``/aihwkit`` (by default, ``~/.config/aihwkit``).
    """

    def __init__(self) -> None:
        self.stored_config = self.parse_config()

    def parse_config(self) -> Dict:
        """Read the configuration from a config file.

        Returns:
            A dictionary with the contents of the ``aihwkit.conf``
            configuration file.
        """
        parser = ConfigParser()
        parser.read(['aihwkit.conf',
                     path.expanduser('{}/aihwkit/aihwkit.conf'.format(
                         getenv('XDG_CONFIG_HOME', '~/.config')))])

        # Check that the expected section is present.
        if 'cloud' in parser:
            return dict(parser['cloud'])

        return {}

    @property
    def token(self) -> str:
        """Return the user token."""
        return getenv('AIHW_API_TOKEN',
                      self.stored_config.get('api_token', None))

    @property
    def url(self) -> str:
        """Return the API URL."""
        return self.stored_config.get('api_url', DEFAULT_URL)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""High level analog tiles (analog)."""

from copy import deepcopy
from typing import Optional, Union, TYPE_CHECKING

from torch import device as torch_device
from torch.cuda import current_device
from torch.cuda import device as cuda_device

from aihwkit.exceptions import CudaError
from aihwkit.simulator.rpu_base import cuda, tiles
from aihwkit.simulator.tiles.base import BaseTile

if TYPE_CHECKING:
    from aihwkit.simulator.configs import (
        InferenceRPUConfig, SingleRPUConfig, UnitCellRPUConfig
    )


class AnalogTile(BaseTile):
    r"""Analog tile.

    This analog tile implements an abstract analog tile where many
    cycle-tp-cycle non-idealities and systematic parameter-spreads
    that can be user-defined.

    In general stochastic bit pulse trains are generate during update
    and device materials (or unit cells) at each cross-point are only
    updated if a coincidence of rows and columns pulses.

    Here, a resistive device material is assumed that response with a
    finite step change of its conductance value that is independent of
    its own conductance value.

    In its basic parameter settings it implements the analog RPU tile
    model described in `Gokmen & Vlasov (2016)`_, but with a number of
    enhancements that are adjustable by parameter settings.

    All tile parameters are given in
    :class:`~aihwkit.simulator.parameters.AnalogTileParameters`.

    **Forward pass**:

    In general, the following analog forward pass is computed:

    .. math::

        \mathbf{y} = f_\text{ADC}((W + \sigma_\text{w}\Xi) \otimes
        (f_\text{DAC}( x/\alpha ) +
        \sigma_\text{inp}\,\boldsymbol{\xi}_1 ) +
        \sigma_\text{out}\,\boldsymbol{\xi}_2)\,s_\alpha\,
        s_\text{out}\,\alpha

    where :math:`W` is the weight matrix, :math:`\mathbf{x}` the input
    vector and the :math:`\Xi,\boldsymbol{\xi}_1,\boldsymbol{\xi}_2`
    Gaussian noise variables (with corresponding matrix and vector
    sizes). The :math:`\alpha` is a scale from the noise management
    (see :data:`rpu_types.NoiseManagementTypeMap`). The symbol
    :math:`\otimes` refers to the 'analog' matrix-vector
    multiplication, that might have additional non-linearities.

    :math:`f_\text{Z}` (with `Z` either `ADC` or `DAC`) indicates the
    discretization to a number of equidistant steps between a bound
    value :math:`-b_\text{Z},\ldots,b_\text{Z}` potentially with
    stochastic rounding (SR):

    .. math::

        f_\text{Z}(x) = \text{round}(x\,
        \frac{r_\text{Z}}{2\,b_\text{Z}} +
        \zeta)\frac{2b_\text{Z}}{r_\text{Z}}

    If SR is enabled :math:`\zeta` is an uniform random :math:`\in
    [-0.5,0.5)`. Otherwise :math:`\zeta=0`.  Inputs are clipped below
    :math:`-b_\text{Z}` and above :math:`b_\text{Z}`

    :math:`r_Z` is the resolution of the `ADC` or `DAC`. E.g. for 8
    bit, it would be :math:`1/256`

    Note:
        Typically the resolution is reduced by 2 level, eg. in case of
        8 bits it is set to :math:`1/254` to account for a
        discretization mirror symmetric around zero, including the zero
        and discarding one value.

    The scalar scale :math:`s_\text{out}` can be set by
    ``out_scale``. The scalar scale :math:`s_\alpha` is an additional
    scale that might be use to map weight better to conductance
    ranges.

    For parameters regarding the forward pass behavior, see
    :class:`~aihwkit.simulator.parameters.AnalogTileInputOutputParameters`.


    **Backward pass**:

    Identical to the forward direction except that the transposed
    weight matrix is used.  Same parameters as during the forward pass
    except that bound management is not supported.

    For parameters regarding the backward pass behavior, see
    :class:`~aihwkit.simulator.parameters.AnalogTileInputOutputParameters`.


    **General weight update**:

    The weight update that theoretically needs to be computed is

    .. math:: w_{ij} = w_{ij} + \lambda d_i\,x_j

    thus the outer product of error vector and input vector.

    Although the update depends on the `ResistiveDevice` used, in
    general, stochastic pulse trains of a given length are drawn,
    where the probability of occurrence of an pulse is proportional to
    :math:`\sqrt{\lambda}d_i` and :math:`\sqrt{\lambda}x_j`
    respectively. Then for each cross-point, in case a coincidence of
    column and row pulses occur, the weight is updated one `step`. For
    details, see `Gokmen & Vlasov (2016)`_.

    The amount of how the weight changes per single step might be
    different for the different resistive devices.

    In pseudo code::

        # generate prob number
        p_i  = quantize(A * d_i, res, sto_round)
        q_j  = quantize(B * x_j, res, sto_round)
        sign = sign(d_i)*sign(x_j)

        # generate pulse trains of length BL
        pulse_train_d = gen_pulse_train(p_i, BL) # e.g 101001001
        pulse_train_x = gen_pulse_train(q_j, BL) # e.g 001010010

        for t in range(BL):
            if (pulse_train_x[t]==1) and (pulse_train_d[t]==1)
                update_once(w_{ij}, direction = sign)

    The probabilities are generated using scaling factors ``A`` and ``B`` that
    are determined by the learning rate and pulse train length ``BL`` (see
    below). ``quantize`` is an optional discretization of the resulting
    probability, to account for limited resolution number in the stochastic
    pulse train generation process on the chip .

    The ``update_once`` functionality is in general dependent on the
    analog tile class.  For `ConstantStep` the step width is
    independent of the actual weight, but has cycle-to-cycle
    variation, device-to-device variation or systematic bias for up
    versus down direction (see below).

    For parameters regarding the update behaviour, see
    :class:`~aihwkit.simulator.parameters.AnalogTileUpdateParameters`.

    Args:
        out_size: output vector size of the tile, ie. the dimension of
            :math:`\mathbf{y}` in case of :math:`\mathbf{y} =
            W\mathbf{x}` (or equivalently the dimension of the
            :math:`\boldsymbol{\delta}` of the backward pass).
        in_size: input vector size, ie. the dimension of the vector
            :math:`\mathbf{x}` in case of :math:`\mathbf{y} =
            W\mathbf{x}`).
        rpu_config: resistive processing unit configuration.
        bias: whether to add a bias column to the tile, ie. :math:`W`
            has an extra column to code the biases. Internally, the
            input :math:`\mathbf{x}` will be automatically expanded by
            an extra dimension which will be set to 1 always.
        in_trans: Whether to assume an transposed input (batch first).
        out_trans: Whether to assume an transposed output (batch first).

    .. _Gokmen & Vlasov (2016): https://www.frontiersin.org/articles/10.3389/fnins.2016.00333/full
    """

    def __init__(
            self,
            out_size: int,
            in_size: int,
            rpu_config: Optional[Union['SingleRPUConfig', 'UnitCellRPUConfig',
                                       'InferenceRPUConfig']] = None,
            bias: bool = False,
            in_trans: bool = False,
            out_trans: bool = False,
    ):
        if not rpu_config:
            # Import `SingleRPUConfig` dynamically to avoid import cycles.
            # pylint: disable=import-outside-toplevel
            from aihwkit.simulator.configs import SingleRPUConfig
            rpu_config = SingleRPUConfig()

        super().__init__(out_size, in_size, rpu_config, bias, in_trans, out_trans)

    def cpu(self) -> 'BaseTile':
        """Return a copy of this tile in CPU memory.

        Note:
            CUDA tiles weight can be accessed by `get_weights` etc
            methods, there is no need to move them to CPU and it is
            currently not supported.

        Returns:
            self in case of CPU

        Raises:
            CudaError: if a CUDA tile is moved to CPU
        """
        if self.is_cuda:
            raise CudaError('Currently it is not possible to move CUDA tile to cpu.')

        return self

    def cuda(
            self,
            device: Optional[Union[torch_device, str, int]] = None
    ) -> 'BaseTile':
        """Return a copy of this tile in CUDA memory.

        Args:
            device: CUDA device

        Returns:
            Self with the underlying C++ tile moved to CUDA memory.

        Raises:
            CudaError: if the library has not been compiled with CUDA.
        """
        if not cuda.is_compiled():
            raise CudaError('aihwkit has not been compiled with CUDA support')

        device = torch_device('cuda', cuda_device(device).idx)

        if self.is_cuda and device != self.device:
            raise CudaError('Cannot switch CUDA devices of existing Cuda tiles')

        if isinstance(self.tile, tiles.AnalogTile):
            with cuda_device(device):
                self.tile = tiles.CudaAnalogTile(self.tile)
                self.is_cuda = True
                self.device = device
                self.analog_ctx.cuda(device)

        return self

    def _create_simulator_tile(
            self,
            x_size: int,
            d_size: int,
            rpu_config: Union['SingleRPUConfig', 'UnitCellRPUConfig', 'InferenceRPUConfig']
    ) -> tiles.AnalogTile:
        """Create a simulator tile.

        Args:
            x_size: input size
            d_size: output size
            rpu_config: resistive processing unit configuration

        Returns:
            a simulator tile based on the specified configuration.
        """
        meta_parameter = rpu_config.as_bindings()
        device_parameter = rpu_config.device.as_bindings()

        return meta_parameter.create_array(x_size, d_size, device_parameter)


class CudaAnalogTile(AnalogTile):
    """Analog tile (CUDA).

    Analog tile that uses GPU for its operation. The instantiation is based on
    an existing non-cuda tile: all the source attributes are copied except
    for the simulator tile, which is recreated using a GPU tile.

    Caution:
        Deprecated. Use `AnalogTile(..).cuda()` instead.

    Args:
        source_tile: tile to be used as the source of this tile
    """

    def __init__(self, source_tile: AnalogTile):
        if not cuda.is_compiled():
            raise CudaError('aihwkit has not been compiled with CUDA support')

        # Create a new instance of the rpu config.
        new_rpu_config = deepcopy(source_tile.rpu_config)

        # Create the tile, replacing the simulator tile.
        super().__init__(source_tile.out_size, source_tile.in_size, new_rpu_config,
                         source_tile.bias, source_tile.in_trans, source_tile.out_trans)

        self.cuda(current_device())



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Device configurations presets for resistive processing units."""

# pylint: disable=too-many-instance-attributes

from typing import List
from dataclasses import dataclass, field
from aihwkit.simulator.configs.utils import IOParameters, UpdateParameters
from aihwkit.simulator.configs.devices import (
    ConstantStepDevice, ExpStepDevice, LinearStepDevice, SoftBoundsDevice, OneSidedUnitCell
)
from aihwkit.simulator.presets.utils import (
    PresetIOParameters, PresetUpdateParameters
)


@dataclass
class ReRamESPresetDevice(ExpStepDevice):
    """Preset configuration for a single RRAM analog resistive processing
    unit based on exp. step device.

    Fit of the model :class:`ExpStepDevice` to  `Gong & al., Nat. Commun., 2018`_.

    .. _`Gong & al., Nat. Commun., 2018`: https://www.nature.com/articles/s41467-018-04485-1
    """
    # pylint: disable=invalid-name

    dw_min: float = 0.00135
    up_down: float = 0.259359

    w_max: float = 1.0
    w_min: float = -1.0

    a: float = -0.5
    b: float = -0.5
    gamma_up: float = 5.
    gamma_down: float = 5.
    A_up: float = -1.18445
    A_down: float = -0.081404

    # Device-to-device var.
    dw_min_dtod: float = 0.2  # a little reduced compared to SB because of non-linearity
    up_down_dtod: float = 0.05

    w_max_dtod: float = 0.3
    w_min_dtod: float = 0.3

    # Cycle-to_cycle.
    dw_min_std: float = 5.0
    write_noise_std: float = 75.0


@dataclass
class ReRamSBPresetDevice(SoftBoundsDevice):
    """Preset configuration for a single ReRAM analog resistive processing
    unit based on soft bounds device.

    Loose fit of the model :class:`SoftBoundsDevice` to  `Gong & al., Nat. Commun., 2018`_.

    Note:
        Here it is assumed that the devices have been calibrated to the symmetry
        point by subtracting a reference device (which is in this case not
        explicitly modeled). For a more accurate fit see :class:`ReRamESPresetDevice`.

    .. _`Gong & al., Nat. Commun., 2018`: https://www.nature.com/articles/s41467-018-04485-1
    """

    dw_min: float = 0.002
    up_down: float = 0.0

    w_max: float = 1.25
    w_min: float = -0.75

    mult_noise: bool = False

    # Device-to-device var.
    dw_min_dtod: float = 0.3
    up_down_dtod: float = 0.01  # assumes symmetry point corrected.

    w_max_dtod: float = 0.3/1.25
    w_min_dtod: float = 0.3/0.75

    # Cycle-to_cycle.
    dw_min_std: float = 3.75
    write_noise_std: float = 56


@dataclass
class CapacitorPresetDevice(LinearStepDevice):
    """Preset configuration for a single capacitor resistive processing
    unit based on linear step device.

    Fit of the model :class:`LinearStepDevice` to  `Li & al., VLSI, 2018`_

    Here some capacitor leakage is assumed as well.

    Caution:
        Capacitor leakage is applied only once per mini-batch and this
        the size of the leakage has to be adapted by the user as it
        depends not only on the size of the leak of the physical
        capacitor but also on the assumptions how much physical time
        is required for a full forward and backward cycle through the
        network (which depends on whether one assumes pipelining or not).

        The parameter ``lifetime`` needs to be adjusted accordingly.

    .. _`Li & al., VLSI, 2018`: https://ieeexplore.ieee.org/abstract/document/8510648
    """

    dw_min: float = 0.005
    up_down: float = 0.0

    w_max: float = 1.0
    w_min: float = -1.0

    mult_noise: bool = False

    gamma_up: float = 0.05  # gamma_up = -slope_up * w_max / dw_min_up
    gamma_down: float = 0.05

    # Device-to-device var.
    dw_min_dtod: float = 0.1
    up_down_dtod: float = 0.06

    w_max_dtod: float = 0.07
    w_min_dtod: float = 0.07

    gamma_up_dtod: float = 0.01
    gamma_down_dtod: float = 0.01

    # Cycle-to_cycle.
    dw_min_std: float = 0.3
    write_noise_std: float = 0.0

    # Slope does not depend on bound.
    mean_bound_reference: bool = True

    # Leakage (in mini-batches).
    lifetime: float = 1.0e6
    lifetime_dtod: float = 0.3


@dataclass
class EcRamPresetDevice(LinearStepDevice):
    """Preset configuration for a single Lithium-based ECRAM resistive processing
    unit based on linear step device.

    Fit of the model :class:`LinearStepDevice` to  `Tang & al., IEDM, 2018`_

    The range is shifted, so that the symmetry point is at zero.

    .. _`Tang & al., IEDM, 2018`: https://ieeexplore.ieee.org/document/8614551
    """

    dw_min: float = 0.002
    up_down: float = 0.0  # assumed shifted onto range

    w_max: float = 1.1724
    w_min: float = -0.8276

    mult_noise: bool = True

    gamma_up: float = 0.1153
    gamma_down: float = 0.5085

    # Device-to-device var.
    dw_min_dtod: float = 0.1
    up_down_dtod: float = 0.01

    w_max_dtod: float = 0.05
    w_min_dtod: float = 0.05

    gamma_up_dtod: float = 0.05
    gamma_down_dtod: float = 0.05

    # Cycle-to_cycle.
    dw_min_std: float = 0.3
    write_noise_std: float = 0.0

    # Slope does not depend on the actual bound.
    mean_bound_reference: bool = True


@dataclass
class EcRamMOPresetDevice(LinearStepDevice):
    """Preset configuration for a single metal-oxide ECRAM resistive processing
    unit based on linear step device.

    Based on data from `Kim et al. IEDM, 2019`_

    .. _`Kim et al. IEDM, 2019`:  https://ieeexplore.ieee.org/document/8993463
    """

    dw_min: float = 2.8214e-4
    up_down: float = 0.0  # assumed shifted onto range

    w_max: float = 1.1714
    w_min: float = -0.8286

    mult_noise: bool = True

    gamma_up: float = 0.4152
    gamma_down: float = 0.7342

    # Device-to-device var.
    dw_min_dtod: float = 0.1
    up_down_dtod: float = 0.01

    w_max_dtod: float = 0.05
    w_min_dtod: float = 0.05

    gamma_up_dtod: float = 0.05
    gamma_down_dtod: float = 0.05

    # Cycle-to_cycle.
    dw_min_std: float = 2.0
    write_noise_std: float = 0.0

    # Slope does not depend on the actual bound.
    mean_bound_reference: bool = True


@dataclass
class IdealizedPresetDevice(ConstantStepDevice):
    """Preset configuration using an idealized device using
    :class:`ConstantStepDevice`.

    (On average) perfectly symmetric device with 10000 steps.

    Definitions are from the specifications listed in `Gokmen &
    Vlasov, Front. Neurosci. 2016`_ (which includes a number of
    device-to-device and cycle-to-cycle variations, see
    :class:`PulsedDevice`), however, setting the device-to-device
    asymmetry term to zero and increasing the number of states by
    roughly 8 (to 10000 states).

    This is the same device used for
    `Rasch, Gokmen & Haensch, IEEE Design & Test, 2019`_.

    .. _`Gokmen & Vlasov, Front. Neurosci. 2016`: \
        https://www.frontiersin.org/articles/10.3389/fnins.2016.00333/full
    .. _`Rasch, Gokmen & Haensch, IEEE Design & Test, 2019`: https://arxiv.org/abs/1906.02698
    """

    dw_min: float = 0.0002  # Factor 5 smaller steps.
    dw_min_dtod: float = 0.3
    dw_min_std: float = 0.3

    up_down: float = 0.0
    up_down_dtod: float = 0.0  # set to zero, since idealized

    w_max: float = 1.0  # Increased range.
    w_min: float = -1.0

    # Device-to-device of range.
    w_max_dtod: float = 0.3
    w_min_dtod: float = 0.3


@dataclass
class GokmenVlasovPresetDevice(ConstantStepDevice):
    """Preset configuration using :class:`ConstantStepDevice`.

    Definitions are (largely) from the specifications listed in `Gokmen &
    Vlasov, Front. Neurosci. 2016`_ (which includes a number of device-to-device
    and cycle-to-cycle variations, see :class:`PulsedDevice`).

    Note, however, that we use some of the algorithmic optimization of the
    follow-up papers as well and here scale everything to the weight range
    ``-1..1``.

    .. _`Gokmen & Vlasov, Front. Neurosci. 2016`: \
        https://www.frontiersin.org/articles/10.3389/fnins.2016.00333/full
    """

    dw_min: float = 0.0016  # to keep 1200 states
    dw_min_dtod: float = 0.3
    dw_min_std: float = 0.3

    up_down: float = 0.0
    up_down_dtod: float = 0.01

    w_max: float = 1.0  # Increased range, parameter adjusted
    w_min: float = -1.0

    # Device-to-device variation of range
    w_max_dtod: float = 0.3
    w_min_dtod: float = 0.3


@dataclass
class PCMPresetDevice(ExpStepDevice):
    """Preset configuration for a single Phase change memory (PCM) analog
    resistive processing unit based on exponential step device model.

    A PCM device based on :math:`Ge_2Sb_2Te_5` described in
    `Nandakumar et al., Front. Neurosci. 2020`_ by using the
    exponential device model with complex cycle-to-cycle noise (see
    :class:`~aihwkit.simulator.configs.device.ExpStepDevice`)

    Note:
        This is an uni-directional device and thus can only be used as
        a plus-minus pair with refresh (see
        :class:`~PCMPresetUnitCell` which is using this device
        combined as a pair based on
        :class:`~aihwkit.simulator.configs.device.OneSidedUnitCell`).

        When the device is reset, device-to-device and cycle-to-cycle
        variation is assumed.

    .. _`Nandakumar et al., Front. Neurosci. 2020`: \
        https://www.frontiersin.org/articles/10.3389/fnins.2020.00406/full
    """
    # pylint: disable=invalid-name

    dw_min: float = 0.01
    up_down: float = 0.0

    w_max: float = 2.
    w_min: float = 0.

    a: float = -1.0  # scales with w_max
    b: float = 0.0
    gamma_up: float = 2.5
    gamma_down: float = 2.5
    A_up: float = -27.235
    A_down: float = -2.235

    # Device-to-device var.
    dw_min_dtod: float = 0.2
    up_down_dtod: float = 0.05

    w_max_dtod: float = 0.1
    w_min_dtod: float = 0.0

    # Cycle-to_cycle.
    dw_min_std: float = 0.6
    dw_min_std_add: float = 0.042  # dw_min_std/0.6*0.5/20
    dw_min_std_slope: float = 0.108  # dw_min_std/0.6*1.3/20

    write_noise_std: float = 0.0

    # reset behavior
    reset: float = 0.01
    reset_dtod: float = 0.02


@dataclass
class PCMPresetUnitCell(OneSidedUnitCell):
    """A unit cell that is comprised of two uni-directional PCM devices of
    opposite sign (see :class:`~PCMPresetDevice`).

    Check for refresh is performed after each mini-batch update. See
    :class:`~aihwkit.simulator.configs.device.OneSidedUnitCell` for
    details on the refresh implementation.
    """

    unit_cell_devices: List = field(
        default_factory=lambda: [PCMPresetDevice(), PCMPresetDevice()])

    refresh_every: int = 1
    units_in_mbatch: bool = True
    refresh_forward: IOParameters = field(default_factory=PresetIOParameters)
    refresh_update: UpdateParameters = field(
        default_factory=lambda: PresetUpdateParameters(desired_bl=31))
